/*
 @LICENSE@
*/

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h> /* for memset() */
#include <stdlib.h>
#include <xmalloc.h>

#include "block_driver.h"
#include "pci_dma.h"
#include "udelay.h"

#include "trace.h"

int BLOCK_COUNT = 256 ;

#define PRD_EOT 0x80000000
#define MAX_STAT_READ 20

/* 
   References:
   ===========

   [1] Programming Interface for Bus Master IDE_Controller Revision 1.0. Brad Hosler 
   Intel

   [2] T13/1410D, AT Attachment with Packet Interface - 6 (ATA/ATAPI-6). www.t13.org

 */

#include "ide.h"
#include "ide_data.reg.h"
#include "ide_bm.reg.h"
//#include "ide_ctl.reg.h"
#ifdef CMD680
#include "cmd680_chipset.h"
#endif

#include <inttypes.h>

#define D_INT 0x1
#define D_SECTORS 0x2
#define D_NEW_COMMAND 0x4
#define D_EDD 0x8
#define D_IDENT 0x10
#define D_NOTIFY 0x20
#define D_INIT 0x40
//#define VERBOSITY (D_INT | D_SECTORS | D_NEW_COMMAND | D_EDD | D_IDENT | D_NOTIFY | D_INIT)
//#define VERBOSITY (0)
#define VERBOSITY (D_INIT | D_IDENT)

#undef DEBUG
/* Seedy GNUism used in macro */
#define DEBUG(lvl, str, ...) { if ((VERBOSITY) & lvl) UserPrint("ide: " str, ##__VA_ARGS__); }

static int do_command(dide_t *dthis, int bus);

lock_t freelock;
static struct command *freelist = NULL;
static struct command *get_command(void);
static void free_command(struct command *);

int nfrees = 0;
int nallocs = 0;
int nrecycles = 0;

static void *ide_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf);
static void  ide_enable(void *device);
static int ide_read(void *device, int diskid, callback_t callback, 
		    callback_data_t callback_data, block_seq blocks); 
static int ide_write(void *device, int diskid, callback_t callback, 
		     callback_data_t callback_data, block_seq blocks);
static int ide_interrupt(void *device);
static int ide_maxdisc(void *device);
static uint64_t ide_size(void *device, int diskid);
static int ide_blksz(void *device, int diskid);


static struct block_ops ide_ops = {
	ide_setup,
	ide_enable,
	ide_read,
	ide_write,
	ide_interrupt,
	ide_maxdisc,
	ide_size,
	ide_blksz
};
static struct driver driver = 
{
	.name = "IDE",
	.class = block_device,
};

struct ide_command_names ide_cmd_names[] = 
{
	{EXECUTE_DEVICE_DIAGNOSTIC, "EXECUTE_DEVICE_DIAGNOSTIC"},
	{IDENTIFY_DEVICE, "IDENTIFY_DEVICE"},
	{READ_DMA, "READ_DMA"},
	{READ_DMA_EXT, "READ_DMA_EXT"},
	{SET_FEATURES, "SET_FEATURES"},
	{WRITE_DMA, "WRITE_DMA"},
	{WRITE_DMA_EXT, "WRITE_DMA_EXT"},
	{0, NULL},
};

static char *
get_command_name(uint16_t cmd)
{
	int i = 0;
	while( ide_cmd_names[i].cmd )
	{
		if (cmd == ide_cmd_names[i].cmd)
			return ide_cmd_names[i].name;
		i++;
	}
	return "Command not known!";
}

struct driver *
ide_init(void)
{
	driver.b_ops = &ide_ops;
	lock_init(&freelock);

	return &driver;
}


static struct command *
get_command(void)
{
	struct command *cp;
	lock_lock(&freelock);
	if ((cp = freelist)) {
		nrecycles++;
		freelist = cp->next;
		lock_unlock(&freelock);
		return cp;
	}

	lock_unlock(&freelock);

	cp = xmalloc(sizeof *cp);
	nallocs++;
	return cp;
}

void 
free_command(struct command *cp)
{
	lock_lock(&freelock);
	cp->next = freelist;
	freelist = cp;
	lock_unlock(&freelock);

	nfrees++;
}

/*
 * setup_sector: This writes the sector registers so it correctly address the 
 * given sector. It currently only handles sectors up to 24 bit, or 8GB.
 *
 * It also assumes the use of LBA.
 */
static void
setup_sector(dide_t * dthis, int bus, int device, uint64_t sector, 
	     uint64_t sectors, int ext)
{
	union ide_dev ide_dev;
	DEBUG(D_SECTORS, "Setting up sectors\n");

	if (sectors > 256)
		ext = 1;
	if (ext == 0) {
		/* Non extended can only handle 28 bit addressing */
		assert(sector < (1 << 28));
		assert(sectors <= (1 << 8));
	} else {
		assert(sector < ((uint64_t) 1 << 48));
		assert(sectors < (1 << 16));
	}

	if (ext == 1) {
		write_ide_sectors(dthis->data_port[bus], 
				  (union ide_sectors) (uint8_t) 
				  ((sectors >> 8) & 0xff)); 
		write_ide_lba_low(dthis->data_port[bus], 
				  (union ide_lba_low) (uint8_t) 
				  ((sector >> 24) & 0xff)); 
		write_ide_lba_mid(dthis->data_port[bus], 
				  (union ide_lba_mid) (uint8_t) 
				  ((sector >> 32) & 0xff)); 
		write_ide_lba_hi(dthis->data_port[bus], 
				 (union ide_lba_hi) (uint8_t) 
				 ((sector >> 40) & 0xff)); 
	}

	/* We possibly need a memory barrier here */
	/* outb 2 */
	write_ide_sectors(dthis->data_port[bus], 
			  (union ide_sectors) (uint8_t) (sectors & 0xff)); 
	/* outb 3 */
	write_ide_lba_low(dthis->data_port[bus], 
			  (union ide_lba_low) (uint8_t) (sector & 0xff)); 
	/* outb 4 */
	write_ide_lba_mid(dthis->data_port[bus], 
			  (union ide_lba_mid) (uint8_t) ((sector >> 8) 
							 & 0xff)); 
	/* outb 5 */
	write_ide_lba_hi(dthis->data_port[bus], 
			  (union ide_lba_hi) (uint8_t) ((sector >> 16) 
							& 0xff)); 
	if (ext == 0) {
		/* In extended addressing these bits aren't used */
		ide_dev.r.sector_hi =  (sector >> 24) & 0xf;
	}
	ide_dev.r.lba = 1;
	if (device) {
		ide_dev.r.dev = 1;
	} else {
		ide_dev.r.dev = 0;
	}
	/* outb 6 */
	write_ide_dev(dthis->data_port[bus], ide_dev);
}


/*
 *  do_command starts the command from the head of the command queue.
 */

static int
do_command(dide_t *dthis, int bus)
{
	union ide_dev ide_dev;
	union ide_command ide_command;
	union ide_status ide_status;
	union idebm_command idebm_command;
	union idebm_prd idebm_prd;
	int cmd = dthis->head[bus]->cmd;
	dmabuf_handle_t dmabuf;
	int i, j, n;
	int ext = 0;
	int isread;

	/* Init register */
	ide_dev.data = 0;
	idebm_command.data = 0;


	assert(dthis->head[bus] != NULL);

	dthis->wfi[bus] = 1;

	/* outb 1 */
	ide_dev.r.dev = dthis->head[bus]->device;
	write_ide_dev(dthis->data_port[bus], ide_dev);

	switch(cmd) {
	case SET_FEATURES:
	case EXECUTE_DEVICE_DIAGNOSTIC:
	case IDENTIFY_DEVICE:
		/* Nothing special for these commands */
		ide_command.r.cmd = dthis->head[bus]->cmd;
		write_ide_command(dthis->data_port[bus], ide_command);

		break;
	case READ_DMA_EXT:
	case WRITE_DMA_EXT:
		ext = 1; 
		/* Fall through */
	case READ_DMA:
	case WRITE_DMA:
		DEBUG(D_NEW_COMMAND, "XFER_DMA: %d, %d, %" PRIu64 ", %d %d\n", bus, 
		      dthis->head[bus]->device, 
		      dthis->head[bus]->block_number, 
		      dthis->head[bus]->blocks,
		      cmd);

		switch (cmd) {
		case READ_DMA:
		case READ_DMA_EXT:
			isread = 1;
			break;
		default:
			isread = 0;
		}
#ifdef IDE_TRACE
		ide_do_trace( (isread ? IDE_TRACE_READ : IDE_TRACE_WRITE),
			      dthis->head[bus]->block_number,
			      dthis->head[bus]->blocks);
#endif
		ide_command.r.cmd = cmd;
		setup_sector(dthis, bus, dthis->head[bus]->device, 
			     dthis->head[bus]->block_number, 
			     dthis->head[bus]->blocks, ext);
		setup_dma(dthis->dma, dthis->head[bus]->data, 
			  512 * dthis->head[bus]->blocks, &dmabuf, 
			  (isread == 1) ? bus_to_mem : mem_to_bus);
		dthis->head[bus]->dmabuf = dmabuf;

		n = GLUE_DMABUF_NENTS(dmabuf);
		for (i = 0, j = 0; i < n; i++, j++) {
			uintptr_t dmaaddr =  GLUE_DMABUF_ENT_DMAADDR(dmabuf, i);
			uintptr_t len = GLUE_DMABUF_ENT_LEN(dmabuf, i);
			while (len > 65536) {
				dthis->dma_table[j].address_pointer
					= dmaaddr;
				dthis->dma_table[j].size_eot 
					= 65536;
				len -= 65536;
				dmaaddr += 65536;
				DEBUG(D_NEW_COMMAND,"dma_table[%d].dmaaddr = 0x%x, "
				      "len 0x%x\n", j,
				      dthis->dma_table[j].address_pointer, 
				      dthis->dma_table[j].size_eot);
				j++;
			}


			dthis->dma_table[j].address_pointer
				= dmaaddr;
			dthis->dma_table[j].size_eot 
				= len;
			DEBUG(D_NEW_COMMAND,"dma_table[%d].dmaaddr = 0x%x, "
			      "len 0x%x\n", j,
			      dthis->dma_table[j].address_pointer, 
			      dthis->dma_table[j].size_eot);
		}
		dthis->dma_table[j-1].size_eot 
			|= PRD_EOT;
		assert(dthis->head[bus]->blocks < 257);


#if 1 /* needed for PIIX */ /* Why? This is done in setup() */
               idebm_prd.r.addr = GLUE_DMABUF_ENT_DMAADDR(dthis->dma_table_dmabuf, 0);
               WRITE_IDEBM_PRD(0, idebm_prd);
#endif

		/* Because the read_write refers to memory, not
		   not the device, the logic here is inverted */
		if (isread == 1) {
			idebm_command.r.read_write = 1;
		} else {
			idebm_command.r.read_write = 0;
		}

		/* outb 7 */
		WRITE_IDEBM_COMMAND(bus, idebm_command);

		/* outb 8 */
		write_ide_command(dthis->data_port[bus], ide_command);

		/* 400ns */
		//udelay(1);

		/* inb 3 (11) */
		ide_status = read_ide_status(dthis->data_port[bus]);

		DEBUG(D_NEW_COMMAND, "XFER_DMA: status %x\n", 
		      ide_status.data);
		if (VERBOSITY & D_NEW_COMMAND)
			print_ide_status(ide_status);

		idebm_command.r.start_stop = 1;
		/* outb 9 (12) */
		WRITE_IDEBM_COMMAND(bus, idebm_command);

		if (VERBOSITY & D_NEW_COMMAND) {
			union idebm_status idebm_status;
			idebm_status = READ_IDEBM_STATUS(bus);
			print_idebm_status(idebm_status);
		}

		break;

	default:
		UserPrint("Unhandled command: %x\n", dthis->head[bus]->cmd);
	}

	return 1;
}

static struct command *
new_command(dide_t * dthis, uint16_t cmd, int bus, int device, int block_number, 
	    int blocks, enum protocol protocol, mem_handle_t data, 
	    bool(*cont_fn)(struct dide *, int), callback_t callback, 
	    callback_data_t callback_data, uintptr_t key)
{
	struct command * next_command;

	next_command = get_command();
	next_command->cmd = cmd;
	next_command->device = device;
	next_command->block_number = block_number;
	next_command->blocks = blocks;
	next_command->protocol = protocol;
	next_command->cont_fn = cont_fn;
	next_command->data = data;
	next_command->done = 0;
	next_command->next = NULL;
	next_command->key = key;
	next_command->callback = callback;
	next_command->callback_data = callback_data;

	lock_lock(&dthis->cmd_lock[bus]);
	if (dthis->head[bus] == NULL) {
		dthis->tail[bus] = dthis->head[bus] = next_command;
		lock_unlock(&dthis->cmd_lock[bus]);
		/* Need to start the ball rolling, we aren't waiting for interrupts */
		DEBUG(D_NEW_COMMAND, "starting command %s (0x%x)\n", get_command_name(cmd), cmd);
		do_command(dthis, bus);
	} else {
		dthis->tail[bus]->next = next_command;
		dthis->tail[bus] = next_command;
		lock_unlock(&dthis->cmd_lock[bus]);
	}
	return next_command;
}

static bool
ident_cont(dide_t * dthis, int bus)
{
	union ide_status ide_status;
	union ide_dev ide_dev;
	union ide_error ide_error;
	int drive = (bus*MAX_DRIVES_BUS);
	uint16_t * data;
	char model[40];

	DEBUG(D_IDENT, "ident_cont called: %d\n", bus );

	ide_status = read_ide_status(dthis->data_port[bus]);
	ide_dev = read_ide_dev(dthis->data_port[bus]);

	if (ide_dev.r.dev) {
		drive += 1;
	}

	data = (uint16_t *)dthis->ident[drive];

	{
		int i;
		for (i = 0; i < 19; i++) {
			model[2*i + 1] = data[i + 27] & 0xff;
			model[2*i] = (data[i + 27] >> 8) & 0xff;
		}
	}

	if (ide_status.r.err) {
		DEBUG(D_IDENT, "%d:%d - Got an error in ident\n", bus, drive );
		ide_error = read_ide_error(dthis->data_port[bus]);
		free(dthis->ident[drive]);
		dthis->ident[drive] = NULL;
	} else {
		if (data[83]&1024) { /* 48-bit addressing */
			dthis->capacity[drive] = ((uint64_t)data[103] << 48) +
				((uint64_t)data[102] << 32) +
				(data[101] << 16) +
				data[100];
		} else {
			dthis->capacity[drive] = ((uint32_t)data[61]<<16) + data[60];
		}

 		DEBUG(D_IDENT, "%d:%d - Model is: (%p) %s\n", bus, drive,
		      data, model);
 		DEBUG(D_IDENT,"Queue depth: %d\n", data[75]);
 		DEBUG(D_IDENT, "Command sets: %x\n", data[83]);
 		DEBUG(D_IDENT, "UDMA Modes: %x %x\n", data[88] >> 8 & 0xff, 
		      data[88] & 0xff);
 		DEBUG(D_IDENT, "MDMA Modes: %x %x\n", data[63] >> 8 & 0xff, 
		      data[63] & 0xff);
		DEBUG(D_IDENT, "Size: %"PRIu64" sectors; %d-bit addressing\n",
		      dthis->capacity[drive],
		      data[83]&1024 ? 48 : 24);

		dthis->large_support[drive] = data[83]&1024;

		dthis->disks++;
		dthis->ready[drive] = 1;
	}
	return false;
}

static bool
init_diag(dide_t * dthis, int bus)
{
	union ide_error ide_error;
	union ide_dev ide_dev;
	mem_handle_t tmp0;
	//	mem_handle_t tmp1;

	DEBUG(D_NOTIFY, "init diag called: %d\n", bus );

	ide_error = read_ide_error(dthis->data_port[bus]);

	ide_dev.r.dev = 1;
	write_ide_dev(dthis->data_port[bus], ide_dev);

	ide_error = read_ide_error(dthis->data_port[bus]);

	dthis->ident[(bus*MAX_DRIVES_BUS)] = xmalloc(512);
	dthis->ident[(bus*MAX_DRIVES_BUS)+1] = xmalloc(512);

	memset(dthis->ident[(bus*MAX_DRIVES_BUS)], 0, 512);
	memset(dthis->ident[(bus*MAX_DRIVES_BUS) + 1], 0, 512);

	tmp0 = new_mem_handle(dthis->ident[(bus*MAX_DRIVES_BUS)], pagesize);
//	tmp1 = new_mem_handle(dthis->ident[(bus*MAX_DRIVES_BUS)+1], pagesize);

	DEBUG(D_IDENT, "Sending IDENTIFY\n");
	new_command(dthis, IDENTIFY_DEVICE, bus, 0, 0, 0, PIO_IN, tmp0, 
		    ident_cont, NULL, null_callback_data, 0);

//	new_command(dthis, IDENTIFY_DEVICE, bus, 1, 0, 0, PIO_IN, tmp1, 
//		    ident_cont, NULL, null_callback_data, 0);

	return false;
}

static bool
init_features(dide_t * dthis, int bus)
{
	union ide_error ide_error;
	mem_handle_t tmp;

	DEBUG(D_NOTIFY, "init features called: %d\n", bus );
	tmp = new_mem_handle(NULL, 0);

	ide_error = read_ide_error(dthis->data_port[bus]);

	new_command(dthis, EXECUTE_DEVICE_DIAGNOSTIC, bus, 0, 0, 0, EDD, tmp, 
		    init_diag, NULL, null_callback_data, 0);
	return false;
}

/* FIXME: this interface is broken. IDE isn't PCI. ceg */
void *
ide_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{
	dide_t * dthis;
	int i, r;
	union idebm_prd idebm_prd;
	union idebm_status idebm_status;

	dthis = malloc(sizeof(dide_t));

	if (dthis == 0) return 0;
	memset(dthis, 0, sizeof *dthis);

	dthis->classp = &driver;
	dthis->dma = dma;
	dthis->pciconf = pciconf;

	/* Setup the component local data */
	dthis->data_port[0] = spaces.v[0];
	dthis->control_port[0] = spaces.v[1];
	dthis->data_port[1] = spaces.v[2];
	dthis->control_port[1] = spaces.v[3];
	dthis->bmaster_port = spaces.v[4];

	for (i = 0; i < MAX_BUS; i++) {
		dthis->wfi[i] = 0;
		/* Added the ports to the clist */
		lock_init(&dthis->cmd_lock[i]);
		dthis->head[i] = NULL;
		dthis->tail[i] = NULL;
	}

	for (i=0; i < MAX_DRIVES; i++) {
		dthis->ready[i] = 0;
		dthis->ident[i] = NULL;
	}

	dthis->disks = 0;

	r = get_pci_consistent_memory(dma, pagesize, &dthis->dma_table_handle, 
				      &dthis->dma_table_dmabuf);
	if (r != 0) return 0;

	dthis->dma_table = get_addr(dthis->dma_table_handle);
	dthis->dma_table_addr = GLUE_DMABUF_ENT_DMAADDR(dthis->dma_table_dmabuf, 0);
	DEBUG(D_INIT, "dma_table is %p\n", dthis->dma_table);

#if 0
	{
		union ide_control x;
		union ide_status s;

		x.data = 0x8;
		x.r.nien = 1;
		x.r.srst = 1;
		DEBUG(D_INIT, "Writing soft reset to drives\n");
		write_ide_control(dthis->control_port[0], x);
		udelay(5000);
		x.r.nien = 0;
		x.r.srst = 0;
		DEBUG(D_INIT, "Clearing Reset pin, enabling interrupts\n");
		write_ide_control(dthis->control_port[0], x);
		s = read_ide_status(dthis->data_port[0]);	
		print_ide_status(s);
		while (s.r.bsy) {
			extern void sleep(int);
			sleep(10);
			DEBUG(D_INIT, "Still busy... \n");
			print_ide_status(s);
			s = read_ide_status(dthis->data_port[0]);
			if (s.r.err == 1) {
				union ide_error e = read_ide_error(dthis->data_port[0]);
				print_ide_error(e);
			}
		}
	}
#endif

	/* FIXME: We only set up bus 0 */
	idebm_status = READ_IDEBM_STATUS(0);
	if (VERBOSITY & D_INIT)
		print_idebm_status(idebm_status);
	idebm_status.r.drive_0_cap = 1;
	WRITE_IDEBM_STATUS(0, idebm_status);
		
	idebm_prd.r.addr = GLUE_DMABUF_ENT_DMAADDR(dthis->dma_table_dmabuf, 0);
	WRITE_IDEBM_PRD(0, idebm_prd);

#ifdef IDE_TRACE
	DEBUG(D_INIT, "setting up IDE trace ...");
	if (ide_setup_trace(IDE_TRACE_FILENAME) != 0)
		DEBUG(D_INIT, "failed!\n");
	DEBUG(D_INIT, " done\n");
#endif

	return dthis;
}

void
ide_enable(void *device)
{
	dide_t *dthis = (dide_t *)device;
	union ide_sectors ide_sectors;
	union ide_feature ide_feature;
        mem_handle_t unused = new_mem_handle(NULL, 0);
	

	/* FIXME: this is broken. We shouldn't have pciconf in IDE, and we
	 * certainly shouldn't be initialising specific chipsets from IDE.
	 */

#ifdef CMD680
//	if( dthis->pciconf.address.address != NULL )
		cmd680_init(dthis);
#endif


	ide_feature.data = 0x3; /* Sub command */
//	ide_sectors.data = 0x45; /* Argument */ /* UDMA mode 5 */
	ide_sectors.data = 0x41; /* Argument */ /* UDMA mode 4? */

	write_ide_feature(dthis->data_port[0], ide_feature);
	write_ide_sectors(dthis->data_port[0], ide_sectors);

	/* Expect interrupts after here */

	new_command(dthis, SET_FEATURES, 0, 0, 0, 0, NONDATA, unused, 
		    init_features, NULL, null_callback_data, 0);

	new_command(dthis, EXECUTE_DEVICE_DIAGNOSTIC, 1, 0, 0, 0, EDD, unused, 
		    init_diag, NULL, null_callback_data, 0);
}

static int
verify_diskid(dide_t * dthis, int diskid)
{
	if (diskid > MAX_DRIVES) {
		return 0;
	}
	if (dthis->ident[diskid] == NULL) {
		return 0;
	}
	return 1;
}

static int
verify_block_range(dide_t * dthis, int diskid, int block_number, int blocks)
{
	int block_count;
	block_count = ide_size(dthis, diskid);
	if (block_number < 0 || block_number + blocks > block_count) {
		return 0;
	}
	return 1;
}

static bool
call_io_complete(struct dide * dthis, int i)
{
	DEBUG(D_NOTIFY, "Call IO complete\n");

	if (dthis->head[i]->callback != NULL && dthis->head[i]->key != 0) {
		dthis->head[i]->callback(dthis->head[i]->callback_data, 
					 dthis->head[i]->key);
	}

	return false;
}


static int 
bennoide_xfer (dide_t *dthis, int diskid, callback_t callback, 
	       callback_data_t callback_data, block_seq data, int rw)
{
	int rv;
	int curblock, count, cmds = 0;
	int i;
	int cmd;


	/* 
	 * Real logic is required here to determine if we want to use
	 * normal DMA or DMA_EXT protocols. For now lets always use DMA_EXT
	 *
	 * The logic needs to be something like:
	 *
	 * if (we support extended && (sector > 28 bit || sector_count > 8 bit):
	 *   use ext
	 * else
	 *   use normal
	 *
	 *
	 * Here we should also do a check that we have valid sector/count etc
	 * or do we want to rely on the calling program?
	 *
	 */

	for (i = 0; i < data.c; i++) {
		if (data.v[i].block_count == 0 && data.v[i].key != 0) {
			callback(callback_data, data.v[i].key);
		}

		rv = verify_block_range(dthis, diskid, 
					data.v[i].blockid, 
					data.v[i].block_count);
		
		if (rv == 0) { 
			UserPrint("ide: private_read error! "); 
			UserPrint("dthis %p, diskid %d, data.v[%d].blockid = %lu, data.v[%d].block_count = %lu\n",
				  dthis, diskid, i, data.v[i].blockid, i, data.v[i].block_count);
			return 1; /* Error */
		}
		
		/* for now we assume it supports DMA - at some speed */
		for (curblock = data.v[i].blockid; 
		    curblock < (data.v[i].blockid + data.v[i].block_count); 
		    curblock += BLOCK_COUNT) {
			bool(*cont_fn)(struct dide *, int) = NULL;
			count = data.v[i].blockid + data.v[i].block_count - curblock;

			if (count > BLOCK_COUNT) count = BLOCK_COUNT; 

			if (dthis->large_support[diskid] && 
			    ((data.v[i].blockid > (1 << 28)) ||
			     (data.v[i].block_count > (1 << 8)))) {
				if (rw == 1) {
					cmd = WRITE_DMA_EXT;
				} else {
					cmd = READ_DMA_EXT;
				} 
			} else {
				if (rw == 1) {
					cmd = WRITE_DMA;
				} else {
					cmd = READ_DMA;
				}
			}

			if (curblock + BLOCK_COUNT 
			    >= (data.v[i].blockid + data.v[i].block_count)) {
				cont_fn = call_io_complete;
			}
			
			new_command(dthis, cmd, diskid / MAX_DRIVES_BUS, 
				    diskid % MAX_DRIVES_BUS, curblock,
				    count, DMA, data.v[i].handle, cont_fn,
				    callback, callback_data, data.v[i].key);
			handle_increment(&data.v[i].handle, (count * 512));
			cmds++;
		}
	}
	return 0;
}

static int 
ide_read (void *device, int diskid, callback_t callback, 
	  callback_data_t callback_data, block_seq blocks)
{
	dide_t *dthis = (dide_t *)device;
	return bennoide_xfer(dthis, diskid, callback, 
			     callback_data, blocks, 0);
}

static int 
ide_write (void *device, int diskid, callback_t callback, 
	   callback_data_t callback_data, block_seq blocks) 
{
	dide_t *dthis = (dide_t *)device;
	return bennoide_xfer(dthis, diskid, callback, 
			     callback_data, blocks, 1);
}

static int
ide_interrupt (void *device)
{
	dide_t *dthis = (dide_t *)device;
	int count = 0;
	union ide_status ide_status;
	union idebm_status idebm_status;
	int i, j;
	int handled = 0;
	struct command * ide_cmd;
	DEBUG(D_INT, "interrupt occurred: %p\n", dthis);

	for (i = 0; i < MAX_BUS; i++) {

		/* inb 1 */
		idebm_status = READ_IDEBM_STATUS(i);

		//print_idebm_status(idebm_status);

		if (idebm_status.r.interrupt == 1) {
			DEBUG(D_INT, "interrupt occurred on bus %d\n", i);
			WRITE_IDEBM_STATUS(i, idebm_status);
		} else {
			continue;
		}


		count = 0;

		if (dthis->wfi[i] == 0) {
			UserPrint("XXXXX: Got an interrupt on this channel but "
				  "we weren't really expecting it.\n");
			continue;
		}

		ide_cmd = dthis->head[i];
		assert(ide_cmd != NULL);

		DEBUG(D_NOTIFY, "protocol is %d\n", ide_cmd->protocol );

		handled = 1;
		switch (ide_cmd->protocol) {
		case EDD:
			DEBUG(D_EDD, "EDD protcol\n");
			ide_status = read_ide_status(dthis->data_port[i]);

			if (ide_status.r.bsy != 0) continue;

			assert(ide_cmd->cont_fn != NULL);
			dthis->wfi[i] = 0;
			ide_cmd->done = 1;

			break;
		case PIO_IN:
			ide_status = read_ide_status(dthis->data_port[i]);
			if ( (ide_status.r.bsy) != 0) {
				UserPrint("not our irq\n");
				continue; /* We assume this interrupt wasn't for us */
			}
			if (ide_status.r.drq == 0) {
				UserPrint("PIO_IN: %d - reading done "
					  "(could be an error) calling "
					  "cont_fn\n", i);
			} else {
				/* read a block doing PIO */
				for (j=0; j < 128; j++) {
					((uint32_t*)get_addr(ide_cmd->data))[j] = 
						read_ide_data(dthis->
							      data_port[i]).
						data;
				}
				free_mem_handle(&ide_cmd->data);
			}
			/* if (all blocks transferred) callback_fn */

			assert(ide_cmd->cont_fn != NULL);
			dthis->wfi[i] = 0;
			ide_cmd->done = 1; /* check this of course.. */

			break;
		case DMA:
			ide_status = read_ide_status(dthis->data_port[i]);
			if ( ide_status.r.bsy == 0 && ide_status.r.drq == 0) {
				if (ide_status.r.err != 0) {
					UserPrint("DMA: Error! (%d)\n", ide_status.r.err);
				}
			}

			teardown_dma(dthis->dma, &ide_cmd->dmabuf);
 
			dthis->wfi[i] = 0;
			ide_cmd->done = 1; /* check this of course.. */

			break;
		case NONDATA:
			/* inb 2 */
			ide_status = read_ide_status(dthis->data_port[i]);
			if (ide_status.r.bsy != 0)
			{
				DEBUG(D_NOTIFY, "not for us...?\n");
				continue; /* We assume this interrupt wasn't for us */
			}

			dthis->wfi[i] = 0;
			ide_cmd->done = 1;
			break;
		default:
			handled = 1;
			UserPrint("Unexpected protocol: %d\n", ide_cmd->protocol);
			assert(!"Unexpected protocol\n");

		}
		if (ide_cmd->done == 1) {
			if (ide_cmd->cont_fn != NULL) {
				ide_cmd->cont_fn(dthis, i);
			} 

			lock_lock(&dthis->cmd_lock[i]);
			dthis->head[i] = ide_cmd->next;
			lock_unlock(&dthis->cmd_lock[i]);

			free_command(ide_cmd);

			if (dthis->head[i]) {
				assert(dthis->head[i]->next != ide_cmd);
				DEBUG(D_NEW_COMMAND, "starting command %s (0x%x)\n", get_command_name(dthis->head[i]->cmd), dthis->head[i]->cmd);

				do_command(dthis, i);
			}
		}

	}
	
	if (handled != 1) {
		UserPrint("Unhandled IDE irq!\n");
	}

	return 0 ;
}

static int
ide_maxdisc (void *device)
{
	return 3;
}

static uint64_t
ide_size (void *device, int diskid)
{
	dide_t *dthis = (dide_t *) device;
	int v = verify_diskid(dthis, diskid);

	if (v == 0) return 0;

	return dthis->capacity[diskid];
}

static int
ide_blksz(void *device, int diskid)
{
	return 512;
}
