#ifdef IDE_TRACE /* only build this in if requested */

#include "trace.h"

/* HELP TEXT */
#define IDE_TRACE_HELP				\
"> [o file] : place output in specified file\n"	\
"> [s] : start IDE trace\n"			\
"> [q] : stop IDE trace\n"			\
"> [d] : disconnect\n"				\
"> [i] : statistical information\n"		\
"> [?] : this text\n"

// the fd to send the trace to
static int ide_trace_fd = 0;
// if we should actually do the trace
static int ide_trace_on = 0;
// set if we are sending just text over the wire (good for debugging)
static int ide_remote_trace = 0;

// statistics counters
static uint64_t total = 0;
static uint64_t writes = 0;
static uint64_t reads = 0;

void strip(char *buf)
{

	int i;
	for(i=0 ; i < strlen(buf) ; i++)
		switch(buf[i])
		{
		case '\n':
		case '\r':
			buf[i] = '\0';
		}
}

void *ide_trace_thread(void *arg)
{
	int nbytes;
	char buf[BUFSIZ], status[BUFSIZ];
	int s, clientfd, clientlen;

	struct sockaddr_in client,svr;

	UserPrint("Trace thread started\n");
	
	/* create socket */
	s = socket(AF_INET, SOCK_STREAM, 0);
	if(s == -1) {
		UserPrint("Socket Error : %s\n", strerror(errno));
		return NULL;
	}
  
	/* bind server port */
	svr.sin_family = AF_INET;
	svr.sin_addr.s_addr = htonl(INADDR_ANY);
	svr.sin_port = htons(7123);
  
	if(bind(s, (struct sockaddr *) &svr, sizeof(svr))<0) {
		perror("cannot bind port ");
		return NULL;
	}
	
	listen(s,5);

	while ( 1 )
	{
		clientlen = sizeof(client);
		clientfd = accept(s, (struct sockaddr *) &client, &clientlen);
		if(clientfd<0) {
			perror("cannot accept connection ");
			return NULL;
		}
		
		bzero(buf, BUFSIZ);
		while ( (nbytes = recv(clientfd, buf, BUFSIZ, 0)) > 0 )
		{
			strip(buf);
			UserPrint("Got command %s (nbytes is %d)\n", buf, nbytes);
			switch(buf[0])
			{
				/* disconnect */
			case 'd':
				if (ide_trace_on)
				{
					sprintf(status, "> IDE trace still active!\n");
					break;
				}
				close(clientfd);
				continue;
			case 'o':
				sprintf(status, "> IDE trace setup to file : %s\n", &buf[2]);
				ide_trace_fd = open(&buf[2], 
						    O_CREAT|O_TRUNC|O_WRONLY, 
						    S_IRUSR|S_IWUSR);
				if (ide_trace_fd < 0)
				{
					sprintf(status, "Unable to open trace file %s : %s\n", 
						  &buf[2], strerror(errno));
				}			
				break;
			case 's':
				if (!ide_trace_fd)
				{
					ide_trace_fd = clientfd;
					ide_remote_trace = 1;
				}
				if (ide_trace_on)
				{
					sprintf(status, "> IDE trace already started!\n");
					break;
				}
				sprintf(status, "> Starting IDE trace\n");
				ide_trace_on = 1;
				break;
			case 'q':
				if (!ide_trace_on)
				{
					sprintf(status, "> IDE trace not started!\n");
					break;
				}
				sprintf(status, "> Finishing IDE trace\n");
				ide_trace_on = 0;
				/* don't close remote connection as that's the fd we're talking on! */
				if (ide_remote_trace)
					ide_remote_trace = 0;
				else 
					close(ide_trace_fd);
				break;
			case 'i' :
				sprintf(status, "> Reads  %"PRId64" [%g%%]\n"
					"> Writes %"PRId64" [%g%%]\n"
					"> ------\n"
					"> Total  %"PRId64"\n"
					"> [Current trace status is %s]\n",
					reads, ((double)reads/(double)total*100.0),
					writes, ((double)writes/(double)total*100.0),
					total,
					ide_trace_on ? "ACTIVE" : "INACTIVE");
				break;
			case '?':
				sprintf(status, IDE_TRACE_HELP); 
				break;
			default:
				sprintf(status, "> Command %s invalid (? for help)\n", buf);
			}
			send(clientfd, status, strlen(status), 0);
#ifdef DEBUG
			UserPrint("%s", status);
#endif
		}

	}
	return 0;
}

int ide_setup_trace(char *filename) 
{
	pthread_t thread;
	int status;

	status = pthread_create(&thread, NULL, ide_trace_thread, NULL);
	if (status != 0)
	{
		UserPrint("Unable to create trace thread\n");
		return -1;
	}
	return 0;
}

int ide_do_trace(int flags, uint64_t block_number, uint16_t blocks)
{
	char buf[BUFSIZ];
	int nbytes;
	struct ide_trace ide_trace =
		{
			.flags = flags,
			.block_number = block_number,
			.blocks = blocks,
		};
	
	if (!ide_trace_on)
		return 1;

	total++;
	ide_trace.flags == IDE_TRACE_READ ? reads++ : writes++;
	
	sprintf(buf, "[%-5s : %-4"PRId64" - %-4"PRId16"]\n", 
		(flags == IDE_TRACE_READ) ? 
		"read" : "write", block_number, blocks);
	
	if (ide_remote_trace)
		nbytes = send(ide_trace_fd, buf, strlen(buf), 0);
	else
		nbytes = write(ide_trace_fd, &ide_trace, sizeof(ide_trace));
	if ( nbytes < 0 )
	{
		UserPrint("Write trace failed : %s", strerror(errno));
		return -1;
	}
	
	return 0;

}

#endif /*IDE_TRACE*/
