#include <driver_interface.h>
#include <io/io.h>
#include "mpilib/mpi_type.h"
#include "mpilib/mpi.h"
#include "mpilib/mpi_cnfg.h"
#include "mpilib/mpi_fc.h"
#include "mpilib/mpi_init.h"
#include "mpilib/mpi_ioc.h"
#include "mpilib/mpi_lan.h"
#include "mpilib/mpi_targ.h"

/* some definitions to make the FreeBSD code compile with minimal changes */
#define INLINE 		__inline
#define PAGE_SIZE	PAGESIZE
#define DELAY(usec)	udelay(usec)
#define EBUSY		-1
#define ETIMEDOUT	-1
#define EIO		-1
#define ENXIO		-1

#define	SCSI_STATUS_OK			0x00
#define	SCSI_STATUS_CHECK_COND		0x02
#define	SCSI_STATUS_COND_MET		0x04
#define	SCSI_STATUS_BUSY		0x08
#define SCSI_STATUS_INTERMED		0x10
#define SCSI_STATUS_INTERMED_COND_MET	0x14
#define SCSI_STATUS_RESERV_CONFLICT	0x18
#define SCSI_STATUS_CMD_TERMINATED	0x22
#define SCSI_STATUS_QUEUE_FULL		0x28

typedef uint8_t u_int8_t;
typedef uint16_t u_int16_t;
typedef uint32_t u_int32_t;
typedef unsigned char *caddr_t;
typedef dma_addr_t bus_addr_t;

#define bzero(s,n) memset(s,0,n)
#define bcopy(s,d,n) memcpy(d,s,n)
#define panic(msg) do { printf("PANIC: %s\n", msg); while(1); } while(0)

/*
 * Singly-linked List declarations.  (from FreeBSD queue.h)
 */
#define	SLIST_HEAD(name, type)						\
struct name {								\
	struct type *slh_first;	/* first element */			\
}

#define	SLIST_ENTRY(type)						\
struct {								\
	struct type *sle_next;	/* next element */			\
}

/*
 * Singly-linked List functions.  (from FreeBSD queue.h)
 */
#define	SLIST_FIRST(head)	((head)->slh_first)

#define	SLIST_INIT(head) do {						\
	SLIST_FIRST((head)) = NULL;					\
} while (0)

#define	SLIST_INSERT_HEAD(head, elm, field) do {			\
	SLIST_NEXT((elm), field) = SLIST_FIRST((head));			\
	SLIST_FIRST((head)) = (elm);					\
} while (0)

#define	SLIST_NEXT(elm, field)	((elm)->field.sle_next)

#define	SLIST_REMOVE_HEAD(head, field) do {				\
	SLIST_FIRST((head)) = SLIST_NEXT(SLIST_FIRST((head)), field);	\
} while (0)


/* Max MPT Reply we are willing to accept (must be power of 2) */
#define MPT_REPLY_SIZE   	128

#define MPT_MAX_REQUESTS(mpt)	((mpt)->is_fc? 1024 : 256)
#define MPT_REQUEST_AREA 512
#define MPT_SENSE_SIZE    32	/* included in MPT_REQUEST_SIZE */
#define MPT_REQ_MEM_SIZE(mpt)	(MPT_MAX_REQUESTS(mpt) * MPT_REQUEST_AREA)

/*
 * We cannot tell prior to getting IOC facts how big the IOC's request
 * area is. Because of this we cannot tell at compile time how many
 * simple SG elements we can fit within an IOC request prior to having
 * to put in a chain element.
 * 
 * Experimentally we know that the Ultra4 parts have a 96 byte request
 * element size and the Fibre Channel units have a 144 byte request
 * element size. Therefore, if we have 512-32 (== 480) bytes of request
 * area to play with, we have room for between 3 and 5 request sized
 * regions- the first of which is the command  plus a simple SG list,
 * the rest of which are chained continuation SG lists. Given that the
 * normal request we use is 48 bytes w/o the first SG element, we can
 * assume we have 480-48 == 432 bytes to have simple SG elements and/or
 * chain elements. If we assume 32 bit addressing, this works out to
 * 54 SG or chain elements. If we assume 5 chain elements, then we have
 * a maximum of 49 seperate actual SG segments.
 */

#define MPT_SGL_MAX		49

#define MPT_REPLY_PTOV(m, x) 		\
	((void *)(&m->reply[((x << 1) - m->reply_phys)]))

enum mpt_req_state {
    REQ_FREE, REQ_IN_PROGRESS, REQ_TIMEOUT, REQ_ON_CHIP, REQ_DONE
};
typedef struct req_entry {
	u_int16_t	index;		/* Index of this entry */
	void *		req_vbuf;	/* Virtual Address of Entry */
	void *		sense_vbuf;	/* Virtual Address of sense data */
	bus_addr_t	req_pbuf;	/* Physical Address of Entry */
	bus_addr_t	sense_pbuf;	/* Physical Address of sense data */
	dmabuf_handle_t	dmap;		/* DMA map for data buffer */
	SLIST_ENTRY(req_entry) link;	/* Pointer to next in list */
	enum mpt_req_state debug;	/* Debugging */
	u_int32_t	sequence;	/* Sequence Number */
	callback_t	callback;
	callback_data_t	callback_data;

} request_t;

typedef struct mpt_softc {

	struct driver *classp;
	bus_space_t iospace;
	dma_handle_t dma;

	unsigned char verbose;
	unsigned char is_fc;

	/*
	 * IOC Facts
	 */
	u_int16_t		mpt_global_credits;
	u_int16_t		request_frame_size;
#ifdef FREEBSD
	u_int8_t		mpt_max_devices;
	u_int8_t		mpt_max_buses;
#endif

	/*
	 * Port Facts
	 */
	u_int16_t		mpt_ini_id;

	/*
	 * CAM && Software Management
	 */
	request_t *		request_pool;
	SLIST_HEAD(req_queue, req_entry) request_free_list;

	caddr_t			request;
	caddr_t			reply;
	bus_addr_t		request_phys;
	bus_addr_t		reply_phys;

	u_int32_t		sequence;	/* Sequence Number */

} mpt_softc_t;

#define _KERNEL
#include "mpt.h"

static INLINE void mpt_write(mpt_softc_t *, size_t, u_int32_t);
static INLINE u_int32_t mpt_read(mpt_softc_t *, int);

static INLINE void
mpt_write(mpt_softc_t *mpt, size_t offset, u_int32_t val)
{
	io_write_32(mpt->iospace, offset, val);
}

static INLINE u_int32_t
mpt_read(mpt_softc_t *mpt, int offset)
{
	return (io_read_32(mpt->iospace, offset));
}

#define mpt_prt(mpt, format...) do { printf("scsi: " format); printf("\n"); } while(0)
