/* 
 * National Semiconductor DP83820 Ethernet definitions
 *
 * DP83820 is a  single-chip 10/100/1000 Mb/s Ethernet Controller  for the PCI
 * bus. The DP83820 fully implements the V2.2 66 MHz, 64-bit PCI bus interface
 * for host  communications with power management  support. Packet descriptors
 * and data are transferred via bus-mastering, reducing the burden on the host
 * CPU.  The DP83820 can support full duplex 10/100/1000 Mb/s transmission and
 * reception.
 *
 * Copyright 2003 (C) DiSy.
 * Written by Daniel Potts <danielp@cse.unsw.edu.au> 2003/01/28
 *
 */

#ifndef _DP83820_H
#define _DP83820_H

/* 
 * PCI Configuration Registers
 *
 * The DP83820 implements a PCI version 2.2 configuration register space. This
 * allows a PCI BIOS to "soft" configure the DP83820. Software Reset has no
 * effect on configuration registers. Hardware Reset returns all configuration
 * registers to their hardware reset state. For all unused registers, writes
 * are ignored, and reads return 0.
 */

#define CFGID	0x00	/* Configuration Identification Register */
#define CFGCS	0x04	/* Configuration Command and Status Register */
#define CFGRID	0x08	/* Configuration Revision ID Register */
#define CFGLAT	0x0C	/* Configuration Latency Timer Register */
#define CFGIOA	0x10	/* Configuration IO Base Address Register */
#define CFGMA	0x14	/* Configuration Memory Address Register */
#define CFGMA1	0x18	/* Configuration Memory Address High Dword Register */

#define CFGSID	0x2C	/* Configuration Subsystem Identificaton Register */
#define CFGROM	0x30	/* Boot ROM Configuration Register */
#define CAPPTR	0x34	/* Capabilities Pointer Register */
#define CFGINT	0x3C	/* Configuration Interrupt Select Register */
#define PMCAP	0x40	/* Power Management Capabilities Register */
#define PMCS	0x44	/* Power Management Control and Status Register */

/*
 * Operational Register Bit definitions
 *
 * The DP83820 provides the following set of operational registers mapped into
 * PCI memory space or I/O space. Writes to reserved register locations are
 * ignored. Reads to reserved register locations return undefined values. When
 * mapped to I/O space, a 256 byte window allows access to all the Operational
 * Registers (00-FCh). When mapped into PCI memory space, a 4096 byte window
 * is enabled. In addition to access to Operational Registers, the PCI
 * Configuration Registers can be read at addresses 200-2FCh. Other addresses
 * provide aliased access to Operational Registers or read-only PCI
 * Configuration Registers.
 */

#define	CR	0x00	/* Command Register */
#define	CFG	0x04	/* Configuration and Media Status Register */
#define MEAR	0x08	/* MII/EEPROM Access Register */
#define PTSCR	0x0C	/* PCI Test Control Register */
#define	ISR	0x10	/* Interrupt Status Register */
#define	IMR	0x14	/* Interrupt Mask Register */
#define	IER	0x18	/* Interrupt Enable Register */
#define	IHR	0x1C	/* Interrupt Holdoff Register */
#define	TXDP	0x20	/* Transmit Descriptor Pointer Register */
#define	TXDP_HI	0x24	/* Transmit Descriptor Pointer High Dword Register */
#define	TXCFG	0x28	/* Transmit Configuration Register */
#define	GPIOR	0x2C	/* General Purpose I/O Control Register */
#define	RXDP	0x30	/* Receive Descriptor pointer Register */
#define	RXDP_HI	0x34	/* Receive Descriptor pointer High Dword Register */
#define	RXCFG	0x38	/* Receive Configuration Register */
#define	PQCR	0x3C	/* Priority Queue Control Register */
#define	WCSR	0x40	/* Wake Command/Status Register */
#define	PCR	0x44	/* Pause Control/Status Register */
#define	RFCR	0x48	/* Receive Filter/Match Control Register */
#define	RFDR	0x4C	/* Receive Filter/Match Data Register */
#define	BRAR	0x50	/* Boot ROM Address Register */
#define	BRDR	0x54	/* Boot ROM Data Register */
#define	SRR	0x58	/* Silicon Revision Register */
#define	MIBC	0x5C	/* Management Information Base Control Register */

#define RXErroredPkts	0x60	/* MIB Data Reg: Packets with errors count */
#define	RXFCSErrors	0x64	/* MIB Data Reg: Frame check sequence errors */
#define	RXMsdPktErrors 	0x68	/* MIB Data Reg: FIFO overrun count */
#define	RXFAErrors	0x6C	/* MIB Data Reg: Alignment error count */
#define	RXSymbolErrors	0x70	/* MIB Data Reg: Symbol error count */
#define	RXFrameTooLong	0x74	/* MIB Data Reg: Frame too long count */
#define	RXIRLErrors	0x78	/* MIB Data Reg: In range length error count */
#define	RXBadOpcodes	0x7C	/* MIB Data Reg: Packet ppcode not supported */
#define	RXPauseFrames	0x80	/* MIB Data Reg: Pause frame received */
#define	TXPauseFrames	0x84	/* MIB Data Reg: Pause frame transmitted */
#define	TXSQEErrors	0x88	/* MIB Data Reg: loss of collision heartbeat */

#define	TXDP1	0xA0	/* Transmit Descriptor pointer prio 1  Register */
#define	TXDP2	0xA4	/* Transmit Descriptor pointer prio 2  Register */
#define	TXDP3	0xA8	/* Transmit Descriptor pointer prio 3  Register */

#define	RXDP1	0xB0	/* Receive Descriptor pointer prio 1  Register */
#define	RXDP2	0xB4	/* Receive Descriptor pointer prio 2  Register */
#define	RXDP3	0xB8	/* Receive Descriptor pointer prio 3  Register */

#define	VRCR	0xBC	/* VLAN/IP Receive Control Register */
#define	VTCR	0xC0	/* VLAN/IP Transmit Control Register */
#define	VDR	0xC4	/* VLAN Data Register */
#define	CGMR	0xC8	/* CGM Control Register */
#define	CCSR	0xCC	/* Clockrun Control/Status Register */

#define	TBICR	0xE0	/* TBI Control Register */
#define	TBISR	0xE4	/* TBI Status Register */
#define	TANAR	0xE8	/* TBI Auto-Neg Advertisement Register */
#define	TANLPAR	0xEC	/* TBI Auto-Neg Link Partner ability Register */
#define	TANER	0xF0	/* TBI Auto-Neg Expansion Register */
#define	TESR	0xF4	/* TBI Extended Status Register */

/*
 * Bit Register Definitions for above registers
 */

/*
 * CR  - Command Register definitions
 */

#define CR_TXE		0x00000001      /* transmitter enable - bit 0 */
#define CR_TXD		0x00000002      /* transmitter disable - bit 1 */
#define CR_RXE		0x00000004      /* receiver enable - bit 2 */
#define CR_RXD		0x00000008      /* receiver disable - bit 3 */
#define CR_TXR		0x00000010      /* transmitter reset - bit 4 */
#define CR_RXR		0x00000020      /* receiver reset - bit 5 */
					/* bit 6 unused */
#define CR_SWI		0x00000080      /* software interrupt - bit 7 */
#define CR_RST		0x00000100      /* soft reset - bit 8 */
#define CR_TXDP_BASE	0x00000200      /* Tx Priority Queue Base - bit 9 */
#define CR_TXDP_MASK	0x00001E00      /* Tx Priority Queue Mask */
#define CR_TXDP_0	0x00000200      /* Tx Priority Queue 0 - bit 9 */
#define CR_TXDP_1	0x00000400      /* Tx Priority Queue 1 - bit 10 */
#define CR_TXDP_2	0x00000800      /* Tx Priority Queue 2 - bit 11 */
#define CR_TXDP_3	0x00001000      /* Tx Priority Queue 3 - bit 12 */

#define CR_RXDP_BASE	0x00002000      /* Tx Priority Queue Base - bit 13 */
#define CR_RXDP_MASK	0x0001E000      /* Tx Priority Queue Mask */
#define CR_RXDP_0	0x00002000      /* Tx Priority Queue 0 - bit 13 */
#define CR_RXDP_1	0x00004000      /* Tx Priority Queue 1 - bit 14 */
#define CR_RXDP_2	0x00008000      /* Tx Priority Queue 2 - bit 15 */
#define CR_RXDP_3	0x00010000      /* Tx Priority Queue 3 - bit 16 */
					/* bits 17 - 32 unused */

/*  
 * CFG  - Configuration and Media Status Register definitions
 */

#define CFG_BEM      	0x00000001      /* big endian mode */
#define CFG_EXT_125	0x00000002
#define CFG_BROM_DIS 	0x00000004      /* disable boot rom interface */
#define CFG_PESEL    	0x00000008      /* assert SERR on DPERR detection */
#define CFG_EXD      	0x00000010      /* disable excessive deferral errors */
#define CFG_POW      	0x00000020      /* program out-of-window */
#define CFG_SB       	0x00000040      /* Single Backoff */
#define CFG_REQALG   	0x00000080      /* PCI bus request algorithm */
#define CFG_EXTSTS_EN 	0x00000100      /* Extended Status Enable */
#define CFG_PHY_DIS  	0x00000200      /* disable internal PHY */
#define CFG_PHY_RST  	0x00000400      /* reset internal PHY */

#define CFG_M64ADDR	0x00000800
#define CFG_DATA64_EN	0x00001000
#define CFG_PCI64_DET	0x00002000
#define CFG_T64ADDR	0x00004000

#define CFG_MWI_DIS	0x00008000
#define CFG_MRM_DIS	0x00010000


#define CFG_PINT_DUPSTS	0x00100000      /* Enable Phy Int. on Duplex Change */
#define CFG_PINT_LNKSTS	0x00080000      /* Enable Phy Int. on Link Change */
#define CFG_PINT_SPDSTS	0x00040000      /* Enable Phy Int. on Speed Change */

#define CFG_AUTO_1000	0x00200000      /* Enable Auto 1000Mb Operation */
#define CFG_MODE_1000	0x00400000      /* Enable 1000Mb Mode/GMII */
#define CFG_AUTO_DUP	0x00800000      /* Enable Auto Full Duplex */

#define CFG_TBI_EN	0x01000000

#define CFG_DUPSTS	0x10000000      /* duplex status */
#define	CFG_DUPSTS_FULL	0x10000000	/* duplex status full test */
#define CFG_DUPSTS_HALF	0x00000000	/* duplex status half test */

#define CFG_SPDSTS_MASK	0x60000000      /* speed status (mask) */
#define CFG_SPDSTS1	0x40000000
#define CFG_SPDSTS0	0x20000000
#define CFG_SPEED_1000	0x00000000      /* speed 1000Mb test */
#define CFG_SPEED_10	0x20000000      /* speed 10Mb  test */
#define CFG_SPEED_100	0x60000000      /* speed 100Mb test */
#define CFG_LNKSTS      0x80000000      /* link status */


/* 
 * MEAR - MII / EEPROM Access Register definitions
 */

#define MEAR_EEDI	0x00000001      /* EEPROM data in (out data out) */
#define MEAR_EEDO	0x00000002      /* EEPROM data out (our data in) */
#define MEAR_EECLK	0x00000004      /* EEPROM serial clock */
#define MEAR_EESEL	0x00000008      /* EEPROM chip select */
#define MEAR_MDIO	0x00000010      /* MII management data input/output */
#define MEAR_MDDIR	0x00000020      /* MII management data direction */
#define MEAR_MDC	0x00000040      /* MII management data clock */

//#define	MIIread		0x60000000l
//#define	MIIwrite	0x50020000l

/* 
 * PTSCR - PCI Test Control Register definitions
 */

#define PTSCR_EEBIST_FAIL	0x00000001      /* EE BIST fail indication */
#define PTSCR_EEBIST_EN		0x00000002      /* enable EEPROM BIST */
#define PTSCR_EELOAD		0x00000004      /* enable EEPROM load */
#define PTSCR_RBIST_RXFFAIL	0x00000008      /* RX Filter BIST fail (RO) */
#define PTSCR_RBIST_HFFAIL	0x00000010      /* Hash Filter BIST fail (RO) */
#define PTSCR_RBIST_TX0FAIL	0x00000020      /* TX Data FIFO BIST fail (RO) */
#define PTSCR_RBIST_TX1FAIL	0x00000040      /* TX FIFO 1 BIST fail (RO) */
#define PTSCR_RBIST_RX0FAIL	0x00000080      /* RX Data FIFO BIST fail (RO) */
#define PTSCR_RBIST_RX1FAIL	0x00000100      /* RX FIFO 1 BIST fail (RO) */
#define PTSCR_RBIST_DONE	0x00000200      /* RX FIFO 1 BIST fail (RO) */
#define PTSCR_RBIST_EN		0x00000400      /* SRAM BIST Enable */
#define PTSCR_BMTM_EN		0x00001000      /* bus master test mode enable */

/*
 * ISR, IMR definitions (INTR_x) - Interrupt Status and Mask Register definitions
 */

//#define	INTR_MASK	0x00000000	/* Interrupt Mask */
#define INTR_RXOK		0x00000001      /* RX OK interrupt */
#define INTR_RXDESC		0x00000002      /* RX descriptor interrupt */
#define INTR_RXERR		0x00000004      /* RX packet error */
#define INTR_RXEARLY		0x00000008      /* RX Early Threshold */
#define INTR_RXIDLE		0x00000010      /* RX Idle */
#define INTR_RXORN		0x00000020      /* RX Overrun */

#define INTR_TXOK		0x00000040      /* TX OK interrupt */
#define INTR_TXDESC		0x00000080      /* TX descriptor interrupt */
#define INTR_TXERR		0x00000100      /* TX packet error */
#define INTR_TXIDLE		0x00000200      /* TX idle */
#define INTR_TXURN		0x00000400      /* TX underrun */

#define INTR_MIB		0x00000800      /* MIB service */
#define INTR_SWI_INTR		0x00001000      /* software interrupt */
#define INTR_PME		0x00002000      /* power management event */
#define INTR_PHY		0x00004000      /* phy interrupt */
#define INTR_HIBERR		0x00008000      /* hi bit error */
#define INTR_RXSOVR		0x00010000      /* rx status fifo overrun */
#define INTR_RTABT		0x00020000      /* received target abort */
#define INTR_RMABT		0x00040000      /* received system abort */
#define INTR_SSERR		0x00080000      /* signalled system error */
#define INTR_DPERR		0x00100000      /* detected parity error */
#define INTR_RXRCMP		0x00200000      /* receive reset complete */
#define INTR_TXRCMP		0x00400000      /* transmit reset complete */

#define INTR_RXDESC0		0x00800000	/* Rx Descriptor for priority Q 0 */
#define INTR_RXDESC1		0x01000000	/* Rx Descriptor for priority Q 1 */
#define INTR_RXDESC2		0x02000000	/* Rx Descriptor for priority Q 2 */
#define INTR_RXDESC3		0x04000000	/* Rx Descriptor for priority Q 3 */

#define INTR_TXDESC0		0x08000000	/* Tx Descriptor for priority Q 0 */
#define INTR_TXDESC1		0x10000000	/* Tx Descriptor for priority Q 1 */
#define INTR_TXDESC2		0x20000000	/* Tx Descriptor for priority Q 2 */
#define INTR_TXDESC3		0x40000000	/* Tx Descriptor for priority Q 3 */

/*
 * IE , IHR - interrupt bit definitions 
 */

#define	IER_IE			0x00000001	/* Interrupt Enable */
#define IHR_IHCTL      		0x00000100	/* Interrupt HoldOff value */

/* 
 * Transmit Descriptor Pointer , Transmit Descriptor Pointer HighWord Register
 */

#define	TXDP_MASK	0xFFFFFFFC	/* Transmit Descriptor Pointer mask */
#define	TXDP_HI_MASK	0xFFFFFFFF	/* Transmit Descriptor High Word
					 * Pointer mask */

/* 
 * TXCFG - Transmit Configuration Register definitions
 */

#define TXCFG_MXDMA8	0x00100000	/* 8 bytes	*/
#define TXCFG_MXDMA16	0x00200000	/* 16 bytes	*/
#define TXCFG_MXDMA32	0x00300000	/* 32 bytes	*/
#define TXCFG_MXDMA64	0x00400000	/* 64 bytes	*/
#define TXCFG_MXDMA128	0x00500000	/* 128 bytes	*/
#define TXCFG_MXDMA256	0x00600000	/* 256 bytes	*/
#define TXCFG_MXDMA512	0x00700000	/* 512 bytes	*/
#define TXCFG_MXDMA1024	0x00000000	/* 1024 bytes	*/

#define TXCFG_ATP	0x10000000	/* Automatic Transmit Padding */
#define TXCFG_HBI	0x40000000	/* Heart Beat Ignore */
#define TXCFG_CSI	0x80000000	/* Carrier Sense Ignore	*/

#define TXCFG_ECRETRY	0x00800000	/* Excessive collision retry enable */

/* 
 * RXCFG - Receive Configuration Register definitions
 */

#define	RXCFG_ALP	0x08000000
#define RXCFG_RX_FD	0x10000000
#define RXCFG_STRIPCRC	0x20000000
#define RXCFG_ARP	0x40000000
#define RXCFG_AEP	0x80000000
#define RXCFG_AIRL	0x04000000

#define RXCFG_MXDMA512	0x00700000	/* 512 bytes	*/

/*
 * PQCR	- Priority Queueing Control Register defintions
 */

#define PQCR_TXPQEN	0x00000001
#define PQCR_RXPQ2	0x00000004
#define PQCR_RXPQ3	0x00000008
#define PQCR_RXPQ4	0x0000000c

/*
 * WCSR - Wake Command / Status Register definitions
 */

#define WCSR_WKPHY	0x00000001
#define WCSR_WKUCP	0x00000002
#define WCSR_WKMCP	0x00000004
#define WCSR_WKBCP	0x00000008
#define WCSR_WKARP	0x00000010
#define WCSR_WKPAT0	0x00000020
#define WCSR_WKPAT1	0x00000040
#define WCSR_WKPAT2	0x00000080
#define WCSR_WKPAT3	0x00000100
#define WCSR_WKMAG	0x00000200
#define WCSR_MPSOE	0x00000400

/*
 * PCR - Pause Control / Status Register definitions
 */

#define PCR_PS_FFLO_2K	0x00040000
#define PCR_PS_FFLO_4K	0x00080000
#define PCR_PS_FFLO_8K	0x000C0000

#define PCR_PS_FFHI_2K	0x00100000
#define PCR_PS_FFHI_4K	0x00200000
#define PCR_PS_FFHI_8K	0x00300000

#define PCR_PS_STLO_2	0x00400000
#define PCR_PS_STLO_4	0x00800000
#define PCR_PS_STLO_8	0x00C00000

#define PCR_PS_STHI_2	0x01000000
#define PCR_PS_STHI_4	0x02000000
#define PCR_PS_STHI_8	0x03000000

#define PCR_PLEN_SEL	0x00010000
#define PCR_PSEN	0x80000000
#define PCR_PS_MCAST	0x40000000

#define PCR_PS_DA	0x20000000
#define PCR_PS_RCVD	0x08000000

/*
 * RFCR -Receive Filter / Match Control Register definitions
 */

#define RFAA_SHIFT		28
#define RFCR_RFEP_SHIFT		16

#define RFCR_ULM		0x00080000      /* U/L bit mask */
#define RFCR_UHEN		0x00100000      /* unicast hash enable */
#define RFCR_MHEN		0x00200000      /* mulitcast hash enable */
#define RFCR_AARP		0x00400000      /* accept ARP packets */

#define RFCR_APAT_MASK		0x07800000	/* (&) accept on pattern
						 * match - bits 26-23 */
#define RFCR_APAT0		0x00800000
#define RFCR_APAT1		0x01000000
#define RFCR_APAT2		0x02000000
#define RFCR_APAT3		0x04000000

#define RFCR_APM		0x08000000	/* accept on perfect match */
#define RFCR_AAU		0x10000000	/* RF accept all unicast */
#define RFCR_AAM		0x20000000	/* RF accept all multicasts */
#define RFCR_AAB		0x40000000	/* RF accept all broadcasts */
#define RFCR_RFEN		0x80000000	/* RF enable */

/*
 * MIBC - Management Information Base Control Register definitions
 */

#define MIBC_FRZ		0x00000002
#define MIBC_ACLR		0x00000004
#define MIBC_MIBS		0x00000008

/*
 * VTCR - VLAN/IP Transmit Control Register definitions
 */

#define VTCR_PPCHK   		0x00000008
#define VTCR_GCHK		0x00000004
#define VTCR_VPPTI   		0x00000002
#define VTCR_VGTI		0x00000001

/*
 * VRCR - VLAN/IP Receive Control Register definitions
 */

#define VRCR_RUDPE		0x00000080
#define VRCR_RTCPE		0x00000040
#define VRCR_RIPE		0x00000020
#define VRCR_IPEN		0x00000010

/* CCSR Register definitions */

#define CCSR_CLKRUN_EN		0x00000001

/* Packet decriptor definitions */

#define CMDSTS_OWN		0x80000000
#define CMDSTS_MORE		0x40000000
#define CMDSTS_INTR		0x20000000
#define CMDSTS_ERR		0x10000000
#define CMDSTS_OK		0x08000000

#define CMDSTS_SIZE_MASK 	0x0000ffff

#endif /* _DP83820_H */
