/************************************************************************
 * dp83820_driver.c -- An ethernet driver for the DP83820
 * See http://www.gelato.unsw.edu.au/IA64wiki/UserLevelDrivers
 */
#include "driver_interface.h"
#include "lock.h"
#include "atomic.h" /* mb() &c */
#include "udelay.h"
#include "netdriver.h"
#include "bfrdriver.h"
#include "assert.h"
#include "pci_dma.h"

#include "dp83820.reg.h"

#include "dp83820_driver.h"
#include "dp83820.h"
#include "dp83820_debug.h"
#include <malloc.h>

#include <string.h> /* for memset */

#define NAME "dp83820"

#include "stats.h"
#undef mb
#define mb()
#define unlikely(cond)	__builtin_expect(cond,0)
#define likely(cond)	__builtin_expect(cond,1)
#define read_itc() ({ unsigned long ret; asm volatile("mov %0 = ar.itc" : "=r"(ret)); ret; })
#define DP83820_MEM_SPACE

//int logf(const char *format, ...);
//#undef PRINT_DEBUG
//#define PRINT_DEBUG(x) logf x

static struct driver dp83820 =
{
	.name = NAME,
	.class = netif_device,
};

static void dp83820_reset( ddp83820_t *dthis);
static int init_receiver ( ddp83820_t *dthis);
static int init_transmitter ( ddp83820_t *dthis);
static int init_transmit_ring( ddp83820_t *dthis);
static int init_receive_ring( ddp83820_t *dthis);
static int init_descriptor_ring( ddp83820_t *dthis,
				  struct dp83820_channel *x,
				  uint32_t ndesc, uint32_t own);
//static void receive_filter_accept_all_packets( ddp83820_t *dthis);
static void receive_filter_accept_my_packets( ddp83820_t *cthis);
static void receive_filter_reject_packets( ddp83820_t *dthis);

static void start_receiver( ddp83820_t *dthis);
static void start_transmitter( ddp83820_t *dthis);
static void eeprom_testload( ddp83820_t *dthis);
static void phy_interrupt( ddp83820_t *dthis );
static inline void handle_tx_irq( ddp83820_t *dthis, union isr isr);
static inline void handle_rx_irq( ddp83820_t *dthis, union isr isr);
static void get_mac_address ( ddp83820_t *dthis);

static int dp83820_getmac(void *void_dthis, unsigned char *buf);

/* Exported functions */
struct driver *dp83820_init(void);
/* The rest will be wrapped via netif_() ?? */

static void
kick_idle_rx(ddp83820_t *dthis)
{
	struct dp83820_channel *rx = &dthis->rx;
	union rxdp rxdp;
	union rxdp_hi rxdp_hi;

	// XXX lock it too

#if 0
	PRINT_DEBUG(("kick_idle_rx: desc_curr %d\n",
		     rx->desc_curr));
#endif

	rx->idle = 0;

	rxdp_hi.data = 0;
	WRITE_RXDP_HI(rxdp_hi);
	rxdp.data = rx->desc_phys
	          + (sizeof(struct dp83820_desc) * rx->desc_curr);
	WRITE_RXDP(rxdp);

	start_receiver(dthis);
}


/*
 * Set up the DP83820; allocate a structure to track this instance, and
 * reset everything
 */
static void *
dp83820_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{
	ddp83820_t *dthis;

	union imr imr;
	union ier ier;
	union cfg cfg;
	union cfg cfg_new;
	union txcfg txcfg;
	union rxcfg rxcfg;
	union vrcr vrcr;
	union vtcr vtcr;

	dthis = malloc(sizeof(ddp83820_t));
	if (dthis == 0)
	    goto out;

	memset(dthis, 0, sizeof (*dthis));

	dthis->classp = &dp83820;
	dthis->dma = dma;
	dthis->pciconf = pciconf;
#ifdef DP83820_MEM_SPACE
	dthis->csr = spaces.v[1];
#else
	dthis->csr = spaces.v[0];
#endif
	lock_init(&dthis->tx.lock);
	lock_init(&dthis->rx.lock);

	printf("CFGCS = %x\n", io_read_16(pciconf, 0x4));
	printf("CFGLAT = %x\n", io_read_8(pciconf, 0xd));
	io_write_16(pciconf, 0x4, io_read_16(pciconf, 0x4) | 0x10 | 0x200);
	io_write_8(pciconf, 0xd, 0xff);
	printf("CFGCS = %x\n", io_read_16(pciconf, 0x4));
	printf("CFGLAT = %x\n", io_read_8(pciconf, 0xd));

	/* disable interrupts */
	imr.data = 0;
	WRITE_IMR(imr);

	ier.data = 0;
	WRITE_IER(ier);
	READ_IER();

	dp83820_reset(dthis);
	eeprom_testload(dthis);

	cfg = READ_CFG();
	cfg_new.data = 0;

	if(cfg.r.pci64_det == 1) {
		cfg_new.r.data64_en = 1;
		UserPrint("dp83820: 64-bit PCI bus detected, good luck.\n");
	} else {
		UserPrint("dp82820: doesn't claim to support 64-bit PCI.\n");
	}

	/*
	 * TODO: inform PCI dma infrastructure of addressable DMA range
	 */

	/* Ensure some sane state */

	//matthewc cfg_new.r.mrm_dis = 1;
	cfg_new.r.pint_dupsts = 1;
	cfg_new.r.pint_lnksts = 1;
	cfg_new.r.pint_spdsts = 1;
	//matthewc cfg_new.r.extsts_en = 1;
	//matthewc cfg_new.r.exd = 1;
	//matthewc cfg_new.r.pesel = 1;
	//matthewc cfg_new.r.reqalg = 1;
	//matthewc cfg_new.r.pow = 1;
	cfg_new.r.mode_1000 = 1; // matthewc
	cfg_new.r.sb = 1; // matthewc

	if (cfg.r.tbi_en) {
		UserPrint("dp83820: Optical transceiver.. not supported.\n");
	}

	WRITE_CFG(cfg_new);

	dthis->cfg = cfg_new.data;
	PRINT_DEBUG(("dp83820: CFG written out is:\n"));
#ifdef DP83820_DEBUG
	print_cfg(cfg_new);
#endif

	/* Linux: Note!  The DMA burst size interacts with packet
	 * transmission, such that the largest packet that
	 * can be transmitted is 8192 - FLTH - burst size.
	 * If only the transmit fifo was larger...
	 */
	txcfg.data = 0;
	txcfg.r.csi = 1;
	txcfg.r.hbi = 1;
	txcfg.r.atp = 1; /* Automatic Transmit Padding */
	txcfg.r.mxdma = 0; /* 1024bytes */
	//txcfg.r.mxdma = (cfg.r.pci64_det == 1) ? 0 : 7;
	/* start PCI fetch when Tx fifo this empty (unit: 32 bytes) */
	txcfg.r.flth = 1024/32 /* 1600 / 32 */;
	/* start sending Ethernet packet when Tx Fifo this full (unit: 32 bytes) */
	txcfg.r.drth = 0;
	WRITE_TXCFG(txcfg);

	/* Flush the interrupt holdoff timer */
	WRITE_IHR((union ihr) (uint32_t) 0);
	//WRITE_IHR((union ihr) (uint32_t) (IHR_IHCTL | 1)); /* 100usec */

	/* Set Rx to full duplex, don't accept runt, errored,
	 * long or length range errored packets, 256 word (1024byte) burst.
	 */
	rxcfg.data = 0;
	//rxcfg.r.aep = 0;
	rxcfg.r.arp = 1; /* Accept Runt Packets */
	rxcfg.r.stripcrc = 1; // matthewc
	rxcfg.r.rx_fd = 1;
	rxcfg.r.alp = 1; /* Accept Long Packets - matthewc */
	//rxcfg.r.airl = 0;
	rxcfg.r.mxdma = 0;
	//rxcfg.r.mxdma = (cfg.r.pci64_det == 1) ? 0 : 7;
                           /* 
			    * 0:1k, 1:8, 2:16; 3:32, 4:64, 5:128, 
			    * 6:256,7:512 bytes 
			    */
	/* start PCI write when Rx fifo this full (unit: 8 bytes) */
	rxcfg.r.drth = 2;
	WRITE_RXCFG(rxcfg);

	vtcr.data = 0;
	//vtcr.r.ppchk = 1;
	WRITE_VTCR(vtcr);

	vrcr.data = 0;
	//vrcr.r.ipen = 1;
	//vrcr.r.ripe = 1;
	//vrcr.r.rtcpe = 1;
	//vrcr.r.rudpe = 1;
	WRITE_VRCR(vrcr);
	/*
	 * Turn off (UNSUPPORTED features, or features we don't want):
	 *   priority queueing
	 *   Pause frames
	 *   Wake On Lan
	 */
	WRITE_PQCR((union pqcr)(uint32_t) 0);
	WRITE_PCR((union pcr)(uint32_t) 0);
	WRITE_WCSR((union wcsr)(uint32_t) 0);

	get_mac_address(dthis);
	UserPrint("dp83820: MAC address %02x:%02x:%02x:%02x:%02x:%02x\n",
		  dthis->mac_address[0], dthis->mac_address[1],
		  dthis->mac_address[2], dthis->mac_address[3],
		  dthis->mac_address[4], dthis->mac_address[5]);

	if(init_receiver(dthis))
	    goto out_free;

	if(init_transmitter(dthis))
	    goto out_free;

	return dthis;

 out_free:
	free(dthis);

 out:
	return NULL;
}

static int
enable(void *dthis_v, struct bfr_drv *bfr_handle)
{
	union ier ier;
	ddp83820_t *dthis = (ddp83820_t *)dthis_v;
	
	if (bfr_register_if(bfr_handle,
			    dthis_v,
			    dthis->rx.desc_cnt,
			    RX_BUF_SIZE,
			    dthis->tx.desc_cnt-2)) {
		UserPrint("Couldn't register interface with buffer manager!\n");
		return 1;
	}

	dthis->buf_man_handle = bfr_handle;

	/* Turn interrupts on */
	ier.data = 1;
	WRITE_IER(ier);
	READ_ISR();

	UserPrint("dp83820: Interrupts enabled.\n");
	return 0;
}

static void
dp83820_cleanup(void *dthis_v)
{
	/* TODO */
}

static void
dp83820_enable_receive(void *dthis_v)
{
	ddp83820_t *dthis = (ddp83820_t *)dthis_v;

	receive_filter_accept_my_packets(dthis);
	kick_idle_rx(dthis);
}


static void
dp83820_interrupt(void *dthis_v)
{
	union isr isr;
	ddp83820_t *dthis = (ddp83820_t *)dthis_v;

	isr = READ_ISR();
	PRINT_DEBUG(("interrupt_event: \n"));

	if (isr.r.hibint || isr.r.txurn || isr.r.rxorn)
		printf("dp83820: isr=%lx\n", isr.data);

#ifdef DP83820_INTERRUPT
	print_isr(isr);
#endif

	if (isr.r.txdesc || isr.r.txidle) {
	    STATS_IF(unsigned long start = read_itc());
	    handle_tx_irq(dthis, isr);
	    STATS_ADD(tx_irq_time, read_itc()-start);
	}

	if (isr.r.rxdesc || isr.r.rxidle) {
	    STATS_IF(unsigned long start = read_itc());
	    handle_rx_irq(dthis, isr);
	    STATS_ADD(rx_irq_time, read_itc()-start);
	}

	if (isr.r.phy) {
	    phy_interrupt(dthis);
	}
}

static inline void
reap_rx_descs(ddp83820_t *dthis, struct dp83820_channel *rx)
{
	struct descriptor *desc;
	uint32_t tmp_bufptr;
	uint32_t tmp_size;
	uintptr_t tmp_key;

	/*
	 * For each packet...
	 */

	desc = &rx->desc[rx->desc_curr].d;
	while ((desc->cmdsts & CMDSTS_OWN) && (desc->bufptr))
	{
	    PRINT_DEBUG(("dp83820: rx receive buf 0x%lx\n",
			     (unsigned long) desc->bufptr));

	    lock_lock(&rx->lock); // just to serialise rx->desc_inuse

	    tmp_bufptr = desc->bufptr;
	    tmp_size = desc->cmdsts & CMDSTS_SIZE_MASK;
	    tmp_key = rx->desc_keys[rx->desc_curr];

	    desc->bufptr = 0;
	    rx->desc_inuse--;
	    if (++(rx->desc_curr) == rx->desc_cnt)
		    rx->desc_curr = 0;

	    lock_unlock(&rx->lock);

	    bfr_feedpacket(dthis->buf_man_handle, tmp_bufptr,
			   tmp_size,
			   tmp_key);
	    
	    desc = &rx->desc[rx->desc_curr].d;
	}
}

static inline void
handle_rx_irq(ddp83820_t *dthis, union isr isr)
{
	struct dp83820_channel *rx = &dthis->rx;
	union rxdp_hi rxdp_hi;
	union rxdp rxdp;

	reap_rx_descs(dthis, rx);

	/*
	 * XXX: it will make sense to move this out shortly (with buff
	 * management)
	 */

	if (isr.r.rxidle) {
		/*
		 * In this case we may have reached an end-of-list in the
		 * card, i.e., the card is out of buffers. We may have freed
		 * some above in RX processing, however the card may need
		 * kicking.
		 *
		 * XXX: may be unnecessary/unsafe to do this unconditionally.
		 */
		lock_lock(&rx->lock);
		if (rx->desc_inuse == 0)
		{
			printf("dp83820: out of rx buffers!\n");
			lock_unlock(&rx->lock);
			PRINT_DEBUG(("dp83820: setting to rx idle, no available desc\n"));
			rx->idle = 1;
		} else {
			printf("dp83820: out of rx buffers (temp)!\n");
			lock_unlock(&rx->lock);

			PRINT_DEBUG(("dp83820: rxidle kick desc_curr %d desc_inuse %d\n",
				     rx->desc_curr, rx->desc_inuse));

			rxdp_hi.data = 0;
			WRITE_RXDP_HI(rxdp_hi);
			rxdp.data = rx->desc_phys
			          + (sizeof(struct dp83820_desc)
			             * rx->desc_curr);
			WRITE_RXDP(rxdp);

			start_receiver(dthis);
		}
	}
}

static inline void
reap_tx_descs(ddp83820_t *dthis, struct dp83820_channel *tx)
{
	struct descriptor *desc;

	desc = &tx->desc[tx->desc_curr].d;

	while (((desc->cmdsts & CMDSTS_OWN) == 0) &&
	       (desc->bufptr))
	{
		PRINT_DEBUG(("dp83820: tx release buf 0x%lx\n",
			     (unsigned long) desc->bufptr));

		lock_unlock(&tx->lock);
		bfr_freepacket(dthis->buf_man_handle,
			       (dma_addr_t) desc->bufptr,
			       tx->desc_keys[tx->desc_curr]);
		lock_lock(&tx->lock);

		desc->bufptr = 0;
		mb();
		tx->desc_inuse--;

		PRINT_DEBUG(("dp83820: tx->desc_inuse is %d\n", tx->desc_inuse));

		if (++(tx->desc_curr) == tx->desc_cnt)
			tx->desc_curr = 0;
		desc = &tx->desc[tx->desc_curr].d;
	}
}

static inline void
handle_tx_irq(ddp83820_t *dthis, union isr isr)
{
	struct dp83820_channel *tx = &dthis->tx;

	PRINT_DEBUG(("tx irq received.\n"));

	lock_lock(&tx->lock);

	if (isr.r.txdesc)
		reap_tx_descs(dthis, tx);

	if (isr.r.txidle)
	{
		if (tx->desc_curr != tx->desc_next_empty)
		{
			tx->idle = 0;
			PRINT_DEBUG(("kicking tx (idle) in irq\n"));
			start_transmitter(dthis);
		} else 	{
			PRINT_DEBUG(("setting tx to idle\n"));
			tx->idle = 1;
		}
	}
	lock_unlock(&tx->lock);
}


/* TODO: redo ring-buffer code */
static int
drv_freepacket(void *dthis_void, dma_addr_t buf_phys, size_t buf_size, uintptr_t key)
{
	ddp83820_t *dthis = (ddp83820_t *)dthis_void;
	struct dp83820_channel *rx = &dthis->rx;
	struct descriptor *desc;
	int r = 1;
	uint32_t i;

	lock_lock(&dthis->rx.lock);

	PRINT_DEBUG(("drv_freepacket: next empty is %d, free %x %x\n",
		     rx->desc_next_empty, buf_phys, buf_size));

	i = rx->desc_next_empty;

	if (rx->desc_inuse < rx->desc_cnt)
	{
		rx->desc_inuse++;
		if (++(rx->desc_next_empty) == rx->desc_cnt)
			rx->desc_next_empty = 0;

		desc = &dthis->rx.desc[i].d;

		if (((desc->cmdsts & CMDSTS_OWN) == 0) || (desc->bufptr != 0))
		{
			assert(!"dp83820: trying to set non empty recv buffer ptr\n");
		}

		dthis->rx.desc_keys[i] = key; // set key from bufmgr

		desc->bufptr = (uint32_t) buf_phys;
		//matthewc desc->extsts = 0;

		mb();
		desc->cmdsts = CMDSTS_INTR | (uint32_t)(buf_size & 0xffffffffUL);
		if(rx->idle) {
		    PRINT_DEBUG(("Kicking rx\n"));
		    kick_idle_rx(dthis);
		}
		r = 0;
	} else
	{
		UserPrint("drv_freepacket when desc_inuse (%u) == desc_cnt (%u)\n",
			  rx->desc_inuse, rx->desc_cnt);
		PRINT_DEBUG(("opps\n"));
		while(1);
	}

	lock_unlock(&dthis->rx.lock);
	return r;
}

static int
drv_feedpacket(void *dthis_void, dma_addr_t buf_phys, size_t buf_length, 
	       int more, enum bfr_packet_type ptype, uintptr_t buf_key)
{
	ddp83820_t *dthis = (ddp83820_t *)dthis_void;
	struct descriptor *desc;
	struct dp83820_channel *tx = &dthis->tx;
	union txdp_hi txdp_hi;
	union txdp txdp;

	lock_lock(&tx->lock);

	PRINT_DEBUG(("dp83820: tx buf_phys %d desc desc_next_empty(used) %d more %d\n",
		     buf_phys, tx->desc_next_empty, more));

	desc = &(tx->desc[tx->desc_next_empty].d);

	if (unlikely(tx->desc_inuse >= (tx->desc_cnt-1)))
	{
		reap_tx_descs(dthis, tx);
		if (tx->desc_inuse >= (tx->desc_cnt-1))
		{
			lock_unlock(&tx->lock);
			return -1;
		}
	}

	tx->desc_keys[tx->desc_next_empty] = buf_key;

	tx->desc_inuse++;
	if (++(tx->desc_next_empty) == tx->desc_cnt)
		tx->desc_next_empty = 0;

	assert(desc->bufptr == 0); /* Sanity check */

	desc->bufptr = (uint32_t)buf_phys;
	//matthewc desc->extsts = card_ptype;
	desc->cmdsts = CMDSTS_OWN |
		(more ? CMDSTS_MORE : CMDSTS_INTR) | (uint32_t)(buf_length & 0xffffffffUL);

	mb();

	/*
	 * Kick card if idle.
	 * sr_curr_offset could be out of date, but we kick it anyway -
	 * it is up to the tx irq handler to update and correct this state.
	 */

	if ((tx->idle == 1) && (more == 0))
	{
		PRINT_DEBUG(("dp83820: kicking idle tx desc_curr %d\n",
			     tx->desc_curr));
		if(READ_CR().r.txe)
			PRINT_DEBUG(("dp83820: XXX TXE already set!\n"));

		tx->idle = 0;
		txdp_hi.data = 0;
		WRITE_TXDP_HI(txdp_hi);
		txdp.data = tx->desc_phys
			+ (sizeof(struct dp83820_desc)
			   * tx->desc_curr);
		WRITE_TXDP(txdp);
		start_transmitter(dthis);
	}
	lock_unlock(&tx->lock);
	return 0;
}




static void
dp83820_reset(ddp83820_t *dthis)
{
	PRINT_DEBUG(("dp83820: resetting chip..\n"));

	union cr cr;


	cr.data = 0;
	cr.r.rst = 1;
	WRITE_CR(cr);

	while (READ_CR().r.rst) {
		/* FIXME: This could spin forever! */
	    udelay(100);
	}
}

static void eeprom_testload(ddp83820_t *dthis)
{
	union ptscr ptscr;
	READ_PTSCR();
	do {
		udelay(100);
	} while (ptscr.r.eebist_en);

	ptscr.data = 0; ptscr.r.eeload = 1;
	WRITE_PTSCR(ptscr);

	do {
	    udelay(100);
	} while (READ_PTSCR().r.eeload);}



static int
init_transmitter( ddp83820_t *dthis)
{
	union imr imr;

	PRINT_DEBUG(("dp83820: Initialising transmitter\n"));

	if(init_transmit_ring(dthis))
	    return 1;

	imr = READ_IMR();
	imr.r.txurn = 1;
	//imr.r.txerr = 1;
	//imr.r.txok = 1;
	imr.r.txidle = 1;
	WRITE_IMR(imr);

	return 0;
}

static int
init_receiver (ddp83820_t *dthis)
{
	union imr imr;
	union ccsr ccsr;

	PRINT_DEBUG(("dp83820: Initialising receiver\n"));

	if(init_receive_ring(dthis))
	    return 1;

	/* clock control/stat */
	ccsr.data = 0; ccsr.r.clkrun_en = 1;
	WRITE_CCSR(ccsr);

	/* Receive filter -- ensure it is off until a client is registered */
	receive_filter_reject_packets(dthis);

	/* Kick the PHY (test the link) */
	phy_interrupt(dthis);

	/* Only enable RX at this stage. */

	imr = READ_IMR();
	imr.r.phy = 1;
	imr.r.rxdesc = 1;
	imr.r.rxidle = 1;
	WRITE_IMR(imr);

	return 0;
}

static void
start_receiver(ddp83820_t *dthis)
{
	union cr cr;
	cr.data = 0; cr.r.rxe = 1;
	WRITE_CR(cr);
}


static void
start_transmitter(ddp83820_t *dthis)
{
	union cr cr;
	cr.data = 0; cr.r.txe = 1;
	WRITE_CR(cr);
}


#if 0
static void
receive_filter_accept_all_packets(ddp83820_t *dthis)
{
	union rfcr rfcr;
	rfcr.data = 0;
	WRITE_RFCR(rfcr);

	rfcr.r.aab = 1;
	rfcr.r.aam = 1;
	rfcr.r.aau = 1;
	rfcr.r.apm = 1;
	rfcr.r.aarp = 1;
	WRITE_RFCR(rfcr);

	rfcr.r.rfen = 1;
	WRITE_RFCR(rfcr);
}
#endif

static void
receive_filter_accept_my_packets(ddp83820_t *dthis)
{
	union rfcr rfcr;
	rfcr.data = 0;
	WRITE_RFCR(rfcr);

//	rfcr.r.aau = 1;		/* Accept all unicast */
	rfcr.r.apm = 1;		/* Accept perfect match.
				 * The perfect match register is set to 
				 * our own MAC address by the eprom on startup
				 */
	rfcr.r.aab = 1;		/* Accept broadcast packets */
	rfcr.r.aarp = 1;	/* Accept all ARP  */
	rfcr.r.rfen = 1;	/* Enable packet filter */
	WRITE_RFCR(rfcr);
}

static void
receive_filter_reject_packets(ddp83820_t *dthis)
{
	union rfcr rfcr;
	rfcr.data = 0;
	WRITE_RFCR(rfcr);
}


static void
get_mac_address(ddp83820_t *dthis)
{
	int      i;
	uint32_t data;
	uint8_t *mac_address = dthis->mac_address;
	union rfcr rfcr;
	rfcr.data = 0;

	/*
	 * Must be done after the EELOAD self test
	 */

	for(i = 0; i < 3; i++) {
		rfcr.r.rfaddr = i*2;
		WRITE_RFCR(rfcr);
		data = READ_RFDR().r.rfdata;
		*mac_address++ = (uint8_t)(data & 0xff);
		*mac_address++ = (uint8_t)((data >> 8) & 0xff);
	}
}

/* Exported function */
static int
dp83820_getmac(void *void_dthis, unsigned char *buf)
{
    int i;
    ddp83820_t *dthis = (ddp83820_t *) void_dthis;

    for(i = 0; i < 6; i++) {
	buf[i] = dthis->mac_address[i];
    }

    return 0;
}


static int
init_descriptor_ring(ddp83820_t *dthis, struct dp83820_channel *x, uint32_t ndesc, uint32_t own)
{
	/* uint32_t pages_to_pin; */
	int i;
	int i_next;
	/* int mem_size = PAGESIZE; */

	x->desc_keys = malloc(ndesc * sizeof (*(x->desc_keys)));
	if(!x->desc_keys) {
		UserPrint(NAME ": Couldn't allocate memory!\n");
		return 1;
	}

	x->desc_cnt = ndesc;
#if 0
	pages_to_pin = ((x->desc_cnt * DESCSIZE) + PAGESIZE -1)/PAGESIZE;
	if (pages_to_pin != 1)
	{
	    UserPrint("dp83820: Recv ring desc page cnt != 1 unsupported.\n");
	    x->desc_cnt = PAGESIZE / DESCSIZE;
	}
#endif

	if (get_pci_consistent_memory(dthis->dma,
				      /* mem_size */ x->desc_cnt * DESCSIZE,
				      &x->mem_handle,
				      &x->dma_handle)) {
	    UserPrint(NAME ": couldn't allocate memory!\n");
	    return 1;
	}

        x->desc = get_addr(x->mem_handle);
        x->desc_phys = GLUE_DMABUF_ENT_DMAADDR(x->dma_handle, 0);

	x->desc_curr = 0;
	x->desc_next_empty = 0;
	x->desc_inuse = 0;

	for (i=0; i < x->desc_cnt; i++)
	{
		i_next = i + 1;
		if (i_next == x->desc_cnt)
			i_next = 0;

		x->desc[i].d.link = x->desc_phys
		                  + (sizeof(struct dp83820_desc) * i_next);
		x->desc[i].d.bufptr = 0;
		x->desc[i].d.cmdsts = own;
		//matthewc x->desc[i].d.extsts = 0;
	}

	return 0;
}

#if 0
static void
destroy_descriptor_ring(ddp83820_t *dthis, struct dp83820_channel *x, uint32_t ndesc, uint32_t own)
{
    /* sjw (16/03/2003): not finished */

    teardown_dma(dthis->dma, &x->dma_handle);
}

#endif

static int
init_transmit_ring(ddp83820_t *dthis)
{
	struct dp83820_channel *tx = &dthis->tx;
	union txdp_hi txdp_hi;
	union txdp txdp;

	if(init_descriptor_ring(dthis, tx, TX_NUMDESC, 0))
	    return 1;

	UserPrint("dp83820: Transmit ring of %u empty descriptors.\n",
		  tx->desc_cnt);

	tx->idle = 1;
	txdp_hi.data = 0;
	WRITE_TXDP_HI(txdp_hi);
	txdp.data = tx->desc_phys
	          + (sizeof(struct dp83820_desc) * tx->desc_curr);
	WRITE_TXDP(txdp);

	return 0;
}

static int
init_receive_ring(ddp83820_t *dthis)
{
	struct dp83820_channel *rx = &dthis->rx;
	union rxdp_hi rxdp_hi;
	union rxdp rxdp;

	if(init_descriptor_ring(dthis, rx, RX_NUMDESC, CMDSTS_OWN))
	    return 1;

	UserPrint("dp83820: Receive ring of %u empty descriptors.\n",
		  rx->desc_cnt);

	rx->idle = 1;
	rxdp_hi.data = 0;
	WRITE_RXDP_HI(rxdp_hi);
	rxdp.data = rx->desc_phys
	          + (sizeof(struct dp83820_desc) * rx->desc_curr);
	WRITE_RXDP(rxdp);

	return 0;
}


static void
phy_interrupt(ddp83820_t *dthis)
{
	static char *speeds[] = { "10", "100", "1000", "1000(?)", "1000F" };
	uint32_t cfg, new_cfg;
	union cfg u_cfg;

	PRINT_DEBUG(("dp83820: Handling copper PHY interrupt.\n"));

	u_cfg = READ_CFG();

	new_cfg = cfg = u_cfg.data;
	cfg ^= (CFG_SPDSTS_MASK | CFG_DUPSTS);

#if 0
	new_cfg &= ~(CFG_SB | CFG_MODE_1000 | CFG_SPDSTS_MASK);
	if (cfg & CFG_SPDSTS1) // Gigabit??
		new_cfg |= CFG_MODE_1000 | CFG_SB;
	else
		new_cfg |= CFG_MODE_1000 | CFG_SB;
		//new_cfg &= ~CFG_MODE_1000 | CFG_SB;

	/* FIXME: write it anyway?? */
	if ((cfg & CFG_LNKSTS) && ((new_cfg ^ dthis->cfg) & CFG_MODE_1000)) {
		u_cfg.data = new_cfg;
		WRITE_CFG(u_cfg);
		dthis->cfg = new_cfg;
	}
#endif

	/* store away new speed status bits */
	dthis->cfg &= ~CFG_SPDSTS_MASK;
	dthis->cfg |= cfg & CFG_SPDSTS_MASK;

	if (cfg & CFG_LNKSTS) {
		UserPrint("dp83820: Link up %s-duplex at %s mbps.\n",
			  (cfg & CFG_DUPSTS) ? "full" : "half",
			  speeds[(cfg/CFG_SPDSTS0) & 3]);
	} else {
		UserPrint("dp83820: Link down.\n");
	}
}

static struct netif_ops nops =
{
	.setup = dp83820_setup,
	.enable = enable,
	.cleanup = dp83820_cleanup,
	.enable_receive = dp83820_enable_receive,
	.interrupt = dp83820_interrupt,
	.feedpacket = drv_feedpacket,
	.freepacket = drv_freepacket,
	.macaddr = dp83820_getmac
};

void dp83820_dump_stats(ddp83820_t *dthis)
{
	printf("RXErroredPkts: %u\n", READ_MIB_RXERROREDPKTS());
	printf("RXFCSErrors: %u\n", READ_MIB_RXFCSERRORS());
	printf("RXMsdPktErrors: %u\n", READ_MIB_RXMSDPKTERRORS());
	printf("RXFAErrors: %u\n", READ_MIB_RXFAERRORS());
	printf("RXSymbolErrors: %u\n", READ_MIB_RXSYMBOLERRORS());
	printf("RXFrameTooLong: %u\n", READ_MIB_RXFRAMETOOLONG());
	printf("RXIRLErrors: %u\n", READ_MIB_RXIRLERRORS());
	printf("RXBadOpcodes: %u\n", READ_MIB_RXBADOPCODES());
	printf("RXPauseFrames: %u\n", READ_MIB_RXPAUSEFRAMES());
	printf("TXPauseFrames: %u\n", READ_MIB_TXPAUSEFRAMES());
	printf("TXSQEErrors: %u\n", READ_MIB_TXSQEERRORS());
}

unsigned int dp83820_get_pending(ddp83820_t *dthis)
{
	struct dp83820_channel *tx = &dthis->tx;
	reap_tx_descs(dthis, tx);
	return tx->desc_inuse;
}

void dp83820_reap_tx_descs(ddp83820_t *dthis)
{
	struct dp83820_channel *tx = &dthis->tx;
	reap_tx_descs(dthis, tx);
}

int dp83820_rx_poll(ddp83820_t *dthis)
{
	struct dp83820_channel *rx = &dthis->rx;
	struct descriptor *desc;

	desc = &rx->desc[rx->desc_curr].d;
	if (!(desc->cmdsts & CMDSTS_OWN))
		return 0;

	reap_rx_descs(dthis, rx);
	return 1;
}

/* XXX vNUMA hack */
void dp83820_enter_virtual(ddp83820_t *dthis)
{
#ifdef DP83820_MEM_SPACE
	uintptr_t base = dthis->csr.address;
	base &= ~(1UL << 63);
	dtr_insert(base, base, 12, (2UL << 9) | (4UL << 2)); /* PTE_AR_RW|PTE_MA_UC */
	dthis->csr.address = base;
#endif
}

struct driver *dp83820_init(void) {
	dp83820.n_ops = &nops;
	return &dp83820;
}
