#ifndef _E1000_H_
#define _E1000_H_

#include <io/io.h>
#include <stdlib.h>
#include "lock.h"
#include "netdriver.h"
#include "pci_dma.h"
#include <assert.h>
#include "udelay.h"
#include <stdbool.h>

struct e1000_adapter;
#include "e1000_hw.h"

#ifdef DBG
#define E1000_DBG(args...) UserPrint("e1000: " args)
#else
#define E1000_DBG(args...)
#endif

#define E1000_ERR(args...) UserPrint("e1000: " args)

#if 0
#define PFX "e1000: "
#define DPRINTK(nlevel, klevel, fmt, args...) \
	(void)((NETIF_MSG_##nlevel & adapter->msg_enable) && \
	UserPrint("%s: %s: " fmt, adapter->netdev->name, \
		__FUNCTION__ , ## args))
#endif
#define E1000_MAX_INTR 1

/* How many descriptors for TX and RX ? */
#define E1000_DEFAULT_TXD                  16
#define E1000_MAX_TXD                      256
#define E1000_MIN_TXD                       10
#define E1000_MAX_82544_TXD               4096
#define E1000_DEFAULT_RXD                  16
#define E1000_MAX_RXD                      256
#define E1000_MIN_RXD                       10
#define E1000_MAX_82544_RXD               4096

/* Supported Rx Buffer Sizes */
#define E1000_RXBUFFER_2048  2048
#define E1000_RXBUFFER_4096  4096
#define E1000_RXBUFFER_8192  8192
#define E1000_RXBUFFER_16384 16384

/* SmartSpeed delimiters */
#define E1000_SMARTSPEED_DOWNSHIFT 3
#define E1000_SMARTSPEED_MAX       15

/* Packet Buffer allocations */
#define E1000_PBA_BYTES_SHIFT 0xA
#define E1000_TX_HEAD_ADDR_SHIFT 7
#define E1000_PBA_TX_MASK 0xFFFF0000

/* Flow Control High-Watermark: 5688 bytes below Rx FIFO size */
#define E1000_FC_HIGH_DIFF 0x1638

/* Flow Control Low-Watermark: 5696 bytes below Rx FIFO size */
#define E1000_FC_LOW_DIFF 0x1640

/* Flow Control Pause Time: 858 usec */
#define E1000_FC_PAUSE_TIME 0x0680

/* How many Tx Descriptors do we need to call netif_wake_queue ? */
#define E1000_TX_QUEUE_WAKE	16
/* How many Rx Buffers do we bundle into one write to the hardware ? */
#define E1000_RX_BUFFER_WRITE	16	/* Must be power of 2 */

#define AUTO_ALL_MODES       0
#define E1000_EEPROM_APME    0x0400

#ifndef E1000_MASTER_SLAVE
/* Switch to override PHY master/slave setting */
#define E1000_MASTER_SLAVE	e1000_ms_hw_default
#endif

/* only works for sizes that are powers of 2 */
#define E1000_ROUNDUP(i, size) ((i) = (((i) + (size) - 1) & ~((size) - 1)))
/* already in e1000_kenge, syt*/
//#define TRUE 0
//#define FALSE 1

/* wrapper around a pointer to a socket buffer,
 * so a DMA handle can be stored along with the buffer */
struct e1000_buffer {
	struct sk_buff *skb;
	uint64_t dma;
	unsigned long length;
	unsigned long time_stamp;
	unsigned int next_to_watch;
};

struct e1000_desc_ring {
	/* pointer to the descriptor ring memory */
	void *desc; 
	/* physical address base of ring memory. syt*/
	uint64_t desc_phys;
	
	/* syt, array of buffer information structs */
	uintptr_t *desc_keys;
	/* meme_handle: reperesent virual address of the descriptor ring, syt */
	mem_handle_t mem_handle;
        /* represent phsical address base od dma memory, syt */
        dmabuf_handle_t dma_handle;
	/* length of descriptor ring in bytes */
	unsigned int size;
	/* number of descriptors in the ring */
	unsigned int count;
	/* next descriptor to associate a buffer with */
	unsigned int next_to_use;
	/* next descriptor to check for DD status bit */
	unsigned int next_to_clean;
	
};

#define E1000_DESC_UNUSED(R) \
	((((R)->next_to_clean > (R)->next_to_use) ? 0 : (R)->count) + \
	(R)->next_to_clean - (R)->next_to_use - 1)

#define E1000_GET_DESC(R, i, type)	(&(((struct type *)((R).desc))[i]))
#define E1000_RX_DESC(R, i)		E1000_GET_DESC(R, i, e1000_rx_desc)
#define E1000_TX_DESC(R, i)		E1000_GET_DESC(R, i, e1000_tx_desc)
#define E1000_CONTEXT_DESC(R, i)	E1000_GET_DESC(R, i, e1000_context_desc)

/* board specific private data structure */

struct e1000_adapter {
	struct driver *classp;
	struct timer_list tx_fifo_stall_timer;
	struct timer_list watchdog_timer;
	struct timer_list phy_info_timer;
	struct vlan_group *vlgrp;
	uint32_t bd_number;
	uint32_t rx_buffer_len;
	uint32_t part_num;
	uint32_t wol;
	uint32_t smartspeed;
	uint32_t en_mng_pt;
	uint16_t link_speed;
	uint16_t link_duplex;
	lock_t stats_lock;
	atomic_t irq_sem;
	struct work_struct tx_timeout_task;
    	uint8_t fc_autoneg;

	struct timer_list blink_timer;
	unsigned long led_status;

	/* TX */
	struct e1000_desc_ring tx_ring;
	uint32_t txd_cmd;
	uint32_t tx_int_delay;
	uint32_t tx_abs_int_delay;
	uint32_t gotcl;
	uint64_t gotcl_old;
	uint64_t tpt_old;
	uint64_t colc_old;
	uint32_t tx_fifo_head;
	uint32_t tx_head_addr;
	uint32_t tx_fifo_size;
	atomic_t tx_fifo_stall;
	bool pcix_82544;
	lock_t tx_lock;
	
	/* RX */
	struct e1000_desc_ring rx_ring;
	uint64_t hw_csum_err;
	uint64_t hw_csum_good;
	uint32_t rx_int_delay;
	uint32_t rx_abs_int_delay;
	bool rx_csum;
	uint32_t gorcl;
	uint64_t gorcl_old;
	lock_t rx_lock;
	
	/* Interrupt Throttle Rate */
	uint32_t itr;

	/* should be placed in module section??, syt*/	
	struct netstats net_stats;

	/* structs defined in e1000_hw.h */
	struct e1000_hw hw;
	struct e1000_hw_stats stats;
	struct e1000_phy_info phy_info;
	struct e1000_phy_stats phy_stats;

	dma_handle_t dma;
	dma_handle_t tx_dma;
	//for buffer manager
	struct bfr_drv *buf_man_handle;
	struct driver *buf_man;
	
	uint8_t mac_address[5];
	
}; 

#endif /* _E1000_H_ */
