#include "driver_interface.h"
#include "netdriver.h"
#include "bfrdriver.h"
#include "pci_dma.h"
#include <malloc.h>

#include "mm.h"

#define NAME "tigon3"

static struct driver tigon3 =
{
	.name = NAME,
	.class = netif_device,
};

typedef struct {
	dma_addr_t mem_base;
	bus_space_t pciconf;
	dma_handle_t dma;
	struct bfr_drv *buf_man_handle;
	LM_TX_PACKET_Q TxPacketPendingQ;
} UM_DEVICE_BLOCK, *PUM_DEVICE_BLOCK;

typedef struct {
	struct driver *classp;
	LM_DEVICE_BLOCK lm;
	UM_DEVICE_BLOCK um;
} tigon3_t;

#define LM_TO_UM_DEVICE(pLMDevice)  ((PUM_DEVICE_BLOCK)((pLMDevice) + 1))

typedef struct {
	void *buffer;
	dma_addr_t paddr;
	unsigned int length;
	uintptr_t key;
} fragment_t;

#define MAX_FRAGMENTS 2
typedef struct {
	fragment_t fragment[2];
} UM_PACKET, *PUM_PACKET;

typedef struct tigon3_packet {
	LM_PACKET lm_packet;
	UM_PACKET mm_packet;
} tigon3_packet_t;

#define LM_TO_UM_PACKET(pLMPacket) ((PUM_PACKET)((pLMPacket) + 1))

int MM_Packet_Desc_Size = sizeof(tigon3_packet_t);
unsigned int packets_out = 0;

LM_STATUS
MM_AllocateSharedMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt, PLM_PHYSICAL_ADDRESS pMemoryBlockPhy,
	LM_BOOL Cached)
{
	mem_handle_t mem_handle;
	dmabuf_handle_t dmabuf_handle;
	UM_DEVICE_BLOCK *pUMDevice = LM_TO_UM_DEVICE(pDevice);

	if (get_pci_consistent_memory(pUMDevice->dma, BlockSize,
				      &mem_handle, &dmabuf_handle) == -1)
		return LM_STATUS_FAILURE;

	*pMemoryBlockVirt = get_addr(mem_handle);
	MM_SetAddr(pMemoryBlockPhy, GLUE_DMABUF_ENT_DMAADDR(dmabuf_handle, 0));
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_AllocateMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
        PLM_VOID *pMemoryBlockVirt)
{
	*pMemoryBlockVirt = malloc(BlockSize);
	return *pMemoryBlockVirt ? LM_STATUS_SUCCESS : LM_STATUS_FAILURE;
}

LM_STATUS
MM_MapMemBase(PLM_DEVICE_BLOCK pDevice)
{
	UM_DEVICE_BLOCK *pUMDevice = LM_TO_UM_DEVICE(pDevice);
	pDevice->pMappedMemBase = (void *)(pUMDevice->mem_base);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_InitializeUmPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUMDevice = LM_TO_UM_DEVICE(pDevice);
	PLM_PACKET pPacket;
	PUM_PACKET pUMPacket;
	void *buffer;
	unsigned int i;

	/* allocate Rx buffers */
	for (i = 0; i < pDevice->RxPacketDescCnt; i++) {
		pPacket = (PLM_PACKET)QQ_PopHead(&pDevice->RxPacketFreeQ.Container);
		if (!pPacket)
			return LM_STATUS_FAILURE;

		buffer = malloc(pPacket->u.Rx.RxBufferSize);
		pUMPacket = LM_TO_UM_PACKET(pPacket);
		pUMPacket->fragment[0].buffer = buffer;
		pUMPacket->fragment[0].paddr = (dma_addr_t)buffer;
		pUMPacket->fragment[0].length = 0;

		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}

	QQ_InitQueue(&pUMDevice->TxPacketPendingQ.Container,MAX_TX_PACKET_DESC_COUNT);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_GetConfig(PLM_DEVICE_BLOCK pDevice)
{
	/* override any device parameters here */
	printf("increasing TxPacketDescCnt (default=%x, new=%x)\n",
			pDevice->TxPacketDescCnt, MAX_TX_PACKET_DESC_COUNT);
	pDevice->TxPacketDescCnt = MAX_TX_PACKET_DESC_COUNT;
#ifdef JUMBO_FRAMES
	pDevice->TxMtu = pDevice->RxMtu = MAX_JUMBO_TX_BUFFER_SIZE;
#endif
	pDevice->RxCoalescingTicks = 1;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateRxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUMDevice = LM_TO_UM_DEVICE(pDevice);
	PLM_PACKET pPacket;
	PUM_PACKET pUMPacket;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->RxPacketReceivedQ.Container);
		if (!pPacket)
			break;

		if (pPacket->PacketStatus != LM_STATUS_SUCCESS) {
			printf("tigon3 Rx: %lx\n", pPacket->PacketStatus);
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			continue;
		}

		pUMPacket = LM_TO_UM_PACKET(pPacket);
		bfr_feedpacket(pUMDevice->buf_man_handle,
				pUMPacket->fragment[0].paddr,
				pPacket->PacketSize,
				(uintptr_t)pPacket);
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_CoalesceTxBuffer(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	/* only needed in certain workaround cases */
	printf("MM_CoalesceTxBuffer not implemented!\n");
	return LM_STATUS_FAILURE;
}

LM_STATUS
MM_IndicateTxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUMDevice = LM_TO_UM_DEVICE(pDevice);
	PLM_PACKET pPacket;
	PUM_PACKET pUMPacket;
	unsigned int frag;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->TxPacketXmittedQ.Container);
		if (!pPacket)
			break;

		pUMPacket = LM_TO_UM_PACKET(pPacket);
		for (frag = 0; frag < pPacket->u.Tx.FragCount; frag++) {
			bfr_freepacket(pUMDevice->buf_man_handle,
					pUMPacket->fragment[frag].paddr,
					pUMPacket->fragment[frag].key);
		}
		pPacket->u.Tx.FragCount = 0;
		QQ_PushTail(&pDevice->TxPacketFreeQ.Container, pPacket);
		packets_out--;
	}

	/* send any pending packets */
	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pUMDevice->TxPacketPendingQ.Container);
		if (!pPacket)
			break;

		pUMPacket = LM_TO_UM_PACKET(pPacket);
		frag = pPacket->u.Tx.FragCount;
		if (MM_ATOMIC_READ(&pDevice->SendBdLeft) < (frag+1))
		{
			QQ_PushHead(&pUMDevice->TxPacketPendingQ.Container, pPacket);
			break;
		}
		printf("sending pending packet\n");
		packets_out++;
		LM_SendPacket(pDevice, pPacket);
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateStatus(PLM_DEVICE_BLOCK pDevice, LM_STATUS Status)
{
	if (Status == LM_STATUS_LINK_DOWN)
		printf("tigon3: link down!\n");
	else if (Status == LM_STATUS_LINK_ACTIVE)
		printf("tigon3: link up!\n");
	return LM_STATUS_SUCCESS;
}

void
MM_MapRxDma(PLM_DEVICE_BLOCK pDevice,
	struct _LM_PACKET *pPacket,
	T3_64BIT_HOST_ADDR *paddr)
{
	UM_PACKET *pUMPacket = LM_TO_UM_PACKET(pPacket);
	/* already mapped, we just return the address */
	MM_SetAddr(paddr, pUMPacket->fragment[0].paddr);
}

void
MM_MapTxDma(PLM_DEVICE_BLOCK pDevice,
	struct _LM_PACKET *pPacket,
	T3_64BIT_HOST_ADDR *paddr,
	LM_UINT32 *len,
	int frag)
{
	UM_PACKET *pUMPacket = LM_TO_UM_PACKET(pPacket);
	/* already mapped, we just return the address */
	MM_SetAddr(paddr, pUMPacket->fragment[frag].paddr);
	*len = pUMPacket->fragment[frag].length;
}

void
MM_UnmapRxDma(LM_DEVICE_BLOCK *pDevice, LM_PACKET *pPacket)
{
	/* nothing to do here */
}

LM_STATUS
MM_FreeRxBuffer(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	/* per-packet opposite of MM_InitializeUmPackets */
	/* we just leave the packets allocated for re-init */
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_Sleep(LM_DEVICE_BLOCK *pDevice, LM_UINT32 msec)
{
	/* FIXME */
	printf("MM_Sleep not implemented!\n");
	return LM_STATUS_SUCCESS;
}

LM_DEVICE_BLOCK *
MM_FindPeerDev(LM_DEVICE_BLOCK *pDevice)
{
	/* only needed in certain workaround cases */
	printf("MM_FindPeerDev not implemented!\n");
	return NULL;
}

LM_STATUS
MM_ReadConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 *pValue32)
{
	UM_DEVICE_BLOCK *pUMDeviceBlock = LM_TO_UM_DEVICE(pDevice);

	//printf("MM_ReadConfig32 %x ", Offset);
	*pValue32 = pci_conf_read_32(pUMDeviceBlock->pciconf, Offset);
	//printf("-> %x\n", *pValue32);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_WriteConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 Value32)
{
	UM_DEVICE_BLOCK *pUMDeviceBlock = LM_TO_UM_DEVICE(pDevice);

	//printf("MM_WriteConfig32 %x %x\n", Offset, Value32);
	pci_conf_write_32(pUMDeviceBlock->pciconf, Offset, Value32);
	return LM_STATUS_SUCCESS;
}



static void *
tigon3_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{
	tigon3_t *dthis;

	dthis = malloc(sizeof(tigon3_t));
	if (!dthis)
		return NULL;

	memset(dthis, 0, sizeof(tigon3_t));
	dthis->classp = &tigon3;
	dthis->um.mem_base = spaces.v[0].address;
	dthis->um.pciconf = pciconf;
	dthis->um.dma = dma;

	printf("Initialising tigon3 @ %lx\n", dthis->um.mem_base);

	if (LM_GetAdapterInfo(&dthis->lm) != LM_STATUS_SUCCESS) {
		printf("tigon3: GetAdapterInfo failed\n");
		free(dthis);
		return NULL;
	}

	if (LM_InitializeAdapter(&dthis->lm) != LM_STATUS_SUCCESS) {
		printf("tigon3: InitializeAdapter failed\n");
		free(dthis);
		return NULL;
	}
	printf("tigon3: init success\n");
	return dthis;
}

static int
tigon3_enable(void *dthis_v, struct bfr_drv *bfr_handle)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;

	printf("tigon3_enable\n");
	if (bfr_register_if(bfr_handle,
			    dthis_v,
			    0,
			    0,
			    0)) {
		printf("Couldn't register interface with buffer manager!\n");
		return 1;
	}
	dthis->um.buf_man_handle = bfr_handle;

	printf("LM_EnableInterrupt\n");
	LM_EnableInterrupt(&dthis->lm);
	return 0;
}

static void
tigon3_cleanup(void *dthis_v)
{
}

static void
tigon3_enable_receive(void *dthis_v)
{
}

static void
tigon3_interrupt(void *dthis_v)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	PLM_DEVICE_BLOCK pLMDevice = &dthis->lm;
#if 0 /* tagged version - why doesn't this work? */
	LM_UINT32 oldtag, newtag;

	MB_REG_WR(pLMDevice, Mailbox.Interrupt[0].Low, 1);
	oldtag = pLMDevice->pStatusBlkVirt->StatusTag;
	printf("oldtag = %x, status = %x\n", oldtag, pLMDevice->pStatusBlkVirt->Status);
	do {
		pLMDevice->pStatusBlkVirt->Status &= ~STATUS_BLOCK_UPDATED;
		LM_ServiceInterrupts(pLMDevice);
		newtag = pLMDevice->pStatusBlkVirt->StatusTag;
		printf("newtag = %x, status = %x\n", newtag, pLMDevice->pStatusBlkVirt->Status);
	} while (newtag != oldtag);
	MB_REG_WR(pLMDevice, Mailbox.Interrupt[0].Low, newtag << 24);
#endif
	do {
		MB_REG_WR(pLMDevice, Mailbox.Interrupt[0].Low, 1);
		pLMDevice->pStatusBlkVirt->Status &= ~STATUS_BLOCK_UPDATED;
		LM_ServiceInterrupts(pLMDevice);
		MB_REG_WR(pLMDevice, Mailbox.Interrupt[0].Low, 0);
	} while (pLMDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED);
}

static int
tigon3_feedpacket(void *dthis_v, dma_addr_t buf_phys, size_t buf_length,
		  int more, enum bfr_packet_type ptype, uintptr_t buf_key)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	PLM_DEVICE_BLOCK pLMDevice = &dthis->lm;
	PUM_DEVICE_BLOCK pUMDevice = &dthis->um;
	PLM_PACKET pLMPacket;
	PUM_PACKET pUMPacket;
	unsigned int frag, i;

	pLMPacket = (PLM_PACKET) QQ_PopHead(&pLMDevice->TxPacketFreeQ.Container);
	if (!pLMPacket) {
		printf("out of Tx buffers!\n");
		return -1;
	}

	frag = pLMPacket->u.Tx.FragCount++;
	if (frag >= MAX_FRAGMENTS) {
		printf("tigon3: too many fragments!\n");
		return -1;
	}

	pUMPacket = LM_TO_UM_PACKET(pLMPacket);
	pUMPacket->fragment[frag].paddr = buf_phys;
	pUMPacket->fragment[frag].length = buf_length;
	pUMPacket->fragment[frag].key = buf_key;
	if (more)
	{
		QQ_PushHead(&pLMDevice->TxPacketFreeQ.Container, pLMPacket);
		return 0;
	}

	if (MM_ATOMIC_READ(&pLMDevice->SendBdLeft) < (frag+1))
	{
		printf("tigon3: out of Tx buffer descriptors, pending packet\n");
		QQ_PushHead(&pUMDevice->TxPacketPendingQ.Container, pLMPacket);
		return 0;
	}

#if 0
		printf("packets_out = %x\n", packets_out);
		do {
			for (i = 0; i < 1000000; i++);
			tigon3_interrupt(dthis_v);
			printf("X");
		} while (MM_ATOMIC_READ(&pLMDevice->SendBdLeft) < (frag+1));
		printf("tigon3: ok!\n");
		printf("packets_out = %x\n", packets_out);
#endif

	packets_out++;
	LM_SendPacket(pLMDevice, pLMPacket);
	return 0;
}

static int
tigon3_freepacket(void *dthis_v, dma_addr_t buf_phys, size_t buf_size, uintptr_t key)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	PLM_DEVICE_BLOCK pLMDevice = &dthis->lm;
	PLM_PACKET pLMPacket = (PLM_PACKET)key;

	QQ_PushTail(&pLMDevice->RxPacketFreeQ.Container, pLMPacket);
	LM_QueueRxPackets(pLMDevice);
	return 0;
}

static int
tigon3_getmac(void *dthis_v, unsigned char *buf)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	int i;

	for (i = 0; i < 6; i++)
		buf[i] = dthis->lm.NodeAddress[i];

	return 0;
}

static int
tigon3_setstatistics(void *dthis_v, struct netstats *nsp)
{
	/* not implemented */
	return 0;
}

static int
tigon3_set_holdoff(void *dthis_v, uint32_t nsec)
{
	/* not implemented */
	return 0;
}

/* XXX vNUMA hack */
void
tigon3_enter_virtual(void *dthis_v)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	uintptr_t base = (uintptr_t)(dthis->lm.pMemView);
	base &= ~(1UL << 63);
	dtr_insert(base, base, 16, (2UL << 9) | (4UL << 2)); /* PTE_AR_RW|PTE_MA_UC */
	dthis->lm.pMemView = (void *)base;
}


unsigned int
tigon3_get_pending(void *dthis_v)
{
	tigon3_t *dthis = (tigon3_t *)dthis_v;
	PLM_DEVICE_BLOCK pLMDevice = &dthis->lm;

	return T3_SEND_RCB_ENTRY_COUNT-1-MM_ATOMIC_READ(&pLMDevice->SendBdLeft);
}


static struct netif_ops nops =
{
	.setup = tigon3_setup,
	.enable = tigon3_enable,
	.cleanup = tigon3_cleanup,
	.enable_receive = tigon3_enable_receive,
	.interrupt = tigon3_interrupt,
	.feedpacket = tigon3_feedpacket,
	.freepacket = tigon3_freepacket,
	.macaddr = tigon3_getmac,
	.setstatistics = tigon3_setstatistics,
	.setholdoff = tigon3_set_holdoff
};

struct driver *tigon3_init(void) {
	tigon3.n_ops = &nops;
	return &tigon3;
}
