#include <driver_interface.h>
#include <malloc.h>
#include "uart16550.reg.h"

#if 1

struct driver uart16550_driver_info = {
	.name = "uart16550",
	/* .class = character_device */
};

struct uart16550 {
	struct driver *classp;
	bus_space_t csr;
	void (*receive_callback)(int c);
};

void *
uart16550_setup(bus_space_t space)
{
	struct uart16550 *dthis;

	dthis = malloc(sizeof(struct uart16550));
	dthis->classp = &uart16550_driver_info;
	dthis->csr = space;
	dthis->receive_callback = NULL;

	/* initialise here? */

	return dthis;
}

void
uart16550_enable_receive(void *dthis_v, void (*callback)(int c))
{
	struct uart16550 *dthis = dthis_v;
	union ier ier;
	union fcr fcr;
	union mcr mcr;

	dthis->receive_callback = callback;

	fcr.data = 0;
	fcr.r.fifo_enable = 1;
	fcr.r.rx_fifo_clear = 1;
	fcr.r.tx_fifo_clear = 1;
	fcr.r.trigger_level = 0; /* 1 char */
	WRITE_FCR(fcr);

	/* OUT2 is master interrupt enable, on some chips at least */
	mcr.data = 0;
	mcr.r.out2 = 1;
	WRITE_MCR(mcr);

	ier.data = 0;
	ier.r.rx_int = 1;
	WRITE_IER(ier);
}

void
uart16550_putchar(void *dthis_v, int c)
{
	struct uart16550 *dthis = dthis_v;
	union data data;

	while (! READ_LSR().r.tx_empty)
		;

	data.data = c;
	WRITE_DATA(data);

	if (c == '\n')
		uart16550_putchar (dthis_v, '\r');
}

int
uart16550_getchar_sync(void *dthis_v)
{
	struct uart16550 *dthis = dthis_v;
	if (READ_LSR().r.rx_avail)
		return READ_DATA().data;
	else
		return 0;
}

void
uart16550_interrupt(void *dthis_v)
{
	struct uart16550 *dthis = dthis_v;
	union iir iir;
	int c;

	static int irq_count;
	if (!(++irq_count % 1000))
		printf("&");

	while (1)
	{
		iir = READ_IIR();
		/* 0 = modem status, 1 = tx, 2 = rx, 3 = line status (from lowest to highest priority) */
		/* only interrupt we have enabled is rx */
		if (iir.r.npending || (iir.r.source != 2))
		{
#if 0
			if (iir.r.npending)
				printf("no interupts pending\n");
			else
				printf("serial irq source=%d\n", iir.r.source);
#endif
			return;
		}

		c = READ_DATA().data;
		if (dthis->receive_callback)
			dthis->receive_callback(c);
	}
}

#else
int virtual_mode = 0;

/* Comport register locations */
#define DATA		0
#define IER		1
#define EIR		2
#define LCR		3
#define MCR		4
#define LSR		5
#define MSR		6
#define DLLO		0
#define DLHI		1

/* LCR related constants */
#define BITS_5          0
#define BITS_6          1
#define BITS_7          2
#define BITS_8          3
#define STOP_ONE	(0 << 2)
#define STOP_TWO	(1 << 2)
#define PARITY_NONE	(0 << 3)
#define PARITY_ODD	(1 << 3)
#define PARITY_EVEN	(3 << 3)
#define PARITY_MARK	(6 << 3)
#define PARITY_SPACE	(7 << 3)
#define BREAK_CONTROL	(1 << 6)
#define DLR_ON		(1 << 7)

/* ISR related constants */
#define RxRDY		(1 << 0)
#define TBE		(1 << 5)

#ifdef ITANIUM2
/* FIXME use HCDP table */
unsigned int com_mmio = 1;
unsigned long com_base = 0xf8030000 /* 0xff5e0000 */;
#else
unsigned int com_mmio = 0;
unsigned long com_base = 0x3f8;
#endif
unsigned long ia64_io_base;

static void out_u8 (unsigned long port, unsigned char data)
{
	unsigned long address = ia64_io_base + (((port >> 2) << 12) | (port & ((1 << 12) - 1)));
	*(volatile unsigned char *)address = data;
}

static unsigned char in_u8 (unsigned long port)
{
	unsigned long address = ia64_io_base + (((port >> 2) << 12) | (port & ((1 << 12) - 1)));
	return *(volatile unsigned char *)address;
}

static void com_out_u8 (unsigned long port, unsigned char data)
{
	if (com_mmio)
		*(unsigned char *)(com_base + port) = data;
	else
		out_u8(com_base + port, data);
}

static unsigned char com_in_u8 (unsigned long port)
{
	if (com_mmio)
		return *(unsigned char *)(com_base + port);
	else
		return in_u8(com_base + port);
}

int putchar (int c)
{
	while (! (com_in_u8 (LSR) & TBE))
		;

	com_out_u8 (DATA, c);

	if (c == '\n')
		putchar ('\r');

	return c;
}

char getc (void)
{
	while (! (com_in_u8 (LSR) & RxRDY))
		;

	return com_in_u8 (DATA);
}

char getc_nonblock (void)
{
	if (! (com_in_u8 (LSR) & RxRDY))
		return 0;

	return com_in_u8 (DATA);
}

void console_init (void)
{
	/* asm("mov %0 = ar.k0" : "=r"(ia64_io_base)); */
#ifdef ITANIUM2
	ia64_io_base = 0x3FFFFFC000000;
#else
	ia64_io_base = 0xFFFFC000000;
#endif
	ia64_io_base |= 1UL << 63;
	com_base |= 1UL << 63;

	/* enable interrupts */
	com_out_u8 (IER, 1);
	com_out_u8 (EIR, 7);
	com_out_u8 (EIR, 1);
	com_out_u8 (MCR, com_in_u8(MCR) | (1 << 3));
}

int serial_interrupt (void)
{
	unsigned char status = com_in_u8(EIR);
	if ((status & 7) == 4)
		return 1;
	return 0;
}

void console_enter_virtual (void)
{
	unsigned long new_ia64_io_base, new_com_base;
	new_ia64_io_base = ia64_io_base & ~(1UL << 63);
	new_com_base = com_base & ~(1UL << 63);
	dtr_insert(new_ia64_io_base, new_ia64_io_base, 26);
#if 0
	pte = (ia64_io_base & ~0xfffUL) | PTE_AR_RW | PTE_D | PTE_A | PTE_MA_UC | PTE_P;
	//asm volatile("mov cr.ifa = %0; mov cr.itir = %1;; itr.d dtr[%2] = %3 ;;" : : "r"(ia64_io_base), "r"(26 << 2), "r"(4), "r"(pte));
#endif
	if (com_mmio)
	{
		dtr_insert(new_com_base, new_com_base, 12);
#if 0
		pte = (com_base & ~0xfffUL) | PTE_AR_RW | PTE_D | PTE_A | PTE_MA_UC | PTE_P;
		//asm volatile("mov cr.ifa = %0; mov cr.itir = %1;; itr.d dtr[%2] = %3 ;;" : : "r"(com_base), "r"(12 << 2), "r"(5), "r"(pte));
#endif
	}
	ia64_io_base = new_ia64_io_base;
	com_base = new_com_base;
	virtual_mode = 1;
}
#endif
