/*
 @LICENSE@
*/

#include <sprintf.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>

/*
 * lookup tables for umaxtostr 
 */
static const char xdigits[16] = "0123456789abcdef";
static const char Xdigits[16] = "0123456789ABCDEF";

/*
 * Convert an unsigned integer to a string of digits in the specified base. 
 * Buf should point to the END of the buffer: 22 characters is probably big
 * enough.  NO '\0' is appended to buf. 
 * 
 * If u == 0, NO digits are generated.  The '0' is supplied by vfprintf using
 * its default zero padding, except in certain rare situations (e.g., "%.0d"). 
 */
static inline char *
umaxtostr(char *buf, uintmax_t u, int base, const char *digits)
{
	unsigned long   u2;

	/*
	 * generate the digits in reverse order 
	 */
#if 0				// UINTMAX_MAX > ULONG_MAX
	/*
	 * Uintmax_t arithmetic may be very slow.  Use it only until the
	 * residual fits in an unsigned long. 
	 */
	while (u > ULONG_MAX) {
		*--buf = digits[u % base];
		u /= base;
	}
#endif
#if 1				// avoid using non-constant div/mod
	switch (base)
	{
		case 8:
			for (u2 = u; u2 != 0UL;) {
				*--buf = digits[u2 % 8];
				u2 /= 8;
			}
			break;
		case 10:
#if 0
			for (u2 = u; u2 != 0UL;) {
				*--buf = digits[u2 % 10];
				u2 /= 10;
			}
			break;
#endif
		case 16:
			for (u2 = u; u2 != 0UL;) {
				*--buf = digits[u2 % 16];
				u2 /= 16;
			}
			break;
	}
#else
	for (u2 = u; u2 != 0UL;) {
		*--buf = digits[u2 % base];
		u2 /= base;
	}
#endif
	return buf;
}

/*
 * Print one formatted field.  The length of s is len; any '\0's in s are
 * IGNORED.  The field may have an optional prefix ('+', ' ', '-', '0x', or
 * '0X', packed into an unsigned int), and is padded appropriately to the
 * specified width.  If width < 0, the field is left-justified. 
 */
static inline int
fprintf1(char *output, const char *s, int len, unsigned int prefix,
	 int prefixlen, int width, int prec)
{
	int             i,
	                r;

	for (r = 0; r < width - prec - prefixlen; r++)
		*output++ = ' ';	/* left-padding (if any) */
	for (; prefix != 0; prefix >>= 8, r++)
		*output++ = prefix & 0377;	/* prefix string */
	for (i = 0; i < prec - len; i++, r++)
		*output++ = '0';	/* zero-padding (if any) */
	for (i = 0; i < len; i++, r++)
		*output++ = s[i];	/* actual string */
	for (; r < -width; r++)
		*output++ = ' ';	/* right-padding (if any) */
	return r;
}

int
sprintf(char *output, const char *fmt, ...)
{
	int             ret;
	va_list         ap;

	va_start(ap, fmt);
	ret = vsprintf(output, fmt, ap);
	va_end(ap);
	return ret;
}

int
snprintf(char *output, size_t size, const char *fmt, ...)
{
	int             ret;
	va_list         ap;

	va_start(ap, fmt);
	ret = vsnprintf(output, size, fmt, ap);
	va_end(ap);
	return ret;
}


int
vsprintf(char *output, const char *fmt, va_list ap)
{
	return vsnprintf(output, -1, fmt, ap);
}

/*
 * parse printf format string 
 */
int
vsnprintf(char *output, size_t n, const char *fmt, va_list ap)
{
	bool            alt,
	                ljust,
	                point,
	                zeropad;
	int             lflags;	/* 'h', 'j', 'l', 't', 'z' */
	unsigned int    prefix;	/* a very small string */
	int             width,
	                prec,
	                base = 0,
	    prefixlen;
	int             r,
	                len;	/* size_t? */
	const char     *p,
	               *s,
	               *digits;
	char            buf[24],
	               *const buf_end = buf + sizeof buf;
	intmax_t        d;
	uintmax_t       u = 0;

	r = 0;
	for (p = fmt; *p != '\0'; p++) {
		if (*p != '%') {
		      putc:
			// putc(*p, f);
			*output++ = *p;
			r++;
			if (r == n)
				goto overflow;
			continue;
		}
		alt = false;
		ljust = false;
		point = false;
		zeropad = false;
		lflags = 0;
		prefix = '\0';
		prefixlen = 0;
		width = 0;
		prec = 1;	/* make sure 0 prints as "0" */
		digits = xdigits;
		for (p++;; p++) {
		      again:
			switch (*p) {
			case '%':
				goto putc;
			case '#':
				alt = true;
				continue;
			case '-':	/* takes precedence over '0' */
				ljust = true;
				continue;
			case '0':
				zeropad = true;
				continue;
			case '+':	/* XXX should take precedence over 
					 * ' ' */
			case ' ':
				prefix = *p;
				prefixlen = 1;
				continue;
			case '*':
				width = va_arg(ap, int);
				if (ljust)
					width = -width;
				continue;
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				/*
				 * width = strtol(p, &p, 10), sort of 
				 */
				width = *p - '0';
				for (p++; (unsigned int) (*p - '0') < 10;
				     p++)
					width = width * 10 + (*p - '0');
				if (ljust)
					width = -width;
				goto again;	/* don't increment p */
			case '.':
				point = true;
				if (*++p == '*') {
					prec = va_arg(ap, int);
					continue;
				} else {
					/*
					 * prec = strtol(p, &p, 10), sort
					 * of 
					 */
					for (prec = 0;
					     (unsigned int) (*p - '0') <
					     10; p++)
						prec =
						    prec * 10 + (*p - '0');
					goto again;	/* don't increment 
							 * p */
				}
			case 'h':
				lflags--;
				continue;
			case 'L':
			case 'l':
				lflags++;
				continue;
			case 't':
			case 'z':
				lflags = 1;	/* assume ptrdiff_t and
						 * size_t are long */
				continue;
			case 'j':
				lflags = 2;	/* assume intmax_t is long 
						 * long */
				continue;
#ifndef	NO_FLOAT
			case 'a':
			case 'A':
			case 'e':
			case 'E':
			case 'f':
			case 'g':
			case 'G':
				/*
				 * NOT IMPLEMENTED 
				 */
				switch (lflags) {
				case 0:
					va_arg(ap, double);
					break;
				case 1:
					va_arg(ap, long double);
					break;
				default:
					goto default_case;
				}
				break;
#endif				/* !NO_FLOAT */
			case 'c':
#ifndef	NO_WCHAR
				/*
				 * NOT IMPLEMENTED 
				 */
				if (lflags > 0)
					va_arg(ap, wchar_t);
				else
#endif
					*(buf_end - 1) = va_arg(ap, int);
				s = buf_end - 1;
				len = 1;
				goto common3;
			case 'd':
			case 'i':
				switch (lflags) {
				case -2:
					// d = va_arg(ap, signed char);
					d = va_arg(ap, int);
					break;
				case -1:
					// d = va_arg(ap, short);
					d = va_arg(ap, int);
					break;
				case 0:
					d = va_arg(ap, int);
					break;
				case 1:
					d = va_arg(ap, long);
					break;
#ifndef	NO_LONG_LONG
				case 2:
					d = va_arg(ap, long long);
					break;
#endif
				default:
					goto default_case;
				}
				if (d < 0LL) {
					/*
					 * safely negate d, even
					 * INTMAX_MIN 
					 */
					u = -(uintmax_t) d;
					prefix = '-';	/* override ' ' or 
							 * '+' */
					prefixlen = 1;
				} else {
					u = d;
				}
				base = 10;
				goto common2;
			case 'n':
				switch (lflags) {
				case -2:
					*va_arg(ap, signed char *) = r;
					break;
				case -1:
					*va_arg(ap, short *) = r;
					break;
				case 0:
					*va_arg(ap, int *) = r;
					break;
				case 1:
					*va_arg(ap, long *) = r;
					break;
#ifndef	NO_LONG_LONG
				case 2:
					*va_arg(ap, long long *) = r;
					break;
#endif
				default:
					goto default_case;
				}
				break;
			case 'o':
				base = 8;
				goto common1;
			case 'p':
				u = (uintptr_t) va_arg(ap, const void *);
				base = 16;
				prec = 2 * sizeof(const void *);
				prefix = '0' | 'x' << 8;
				prefixlen = 2;
				goto common2;
			case 's':
				s = va_arg(ap, const char *);
				/*
				 * XXX left-justified strings are scanned
				 * twice 
				 */
				if (point) {
					/*
					 * len = min(prec, strlen(s)) 
					 */
					for (len = 0; len < prec; len++)
						if (s[len] == '\0')
							break;
				} else {
					len = strlen(s);
				}
				goto common3;
			case 'u':
				base = 10;
				goto common1;
			case 'X':
				digits = Xdigits;
				/*
				 * FALLTHROUGH 
				 */
			case 'x':
				base = 16;
				if (alt) {
					prefix = '0' | *p << 8;
					prefixlen = 2;
				}
				/*
				 * FALLTHROUGH 
				 */
			      common1:
				/*
				 * common code for %o, %u, %X, and %x 
				 */
				switch (lflags) {
				case -2:
					// u = va_arg(ap, unsigned char);
					u = va_arg(ap, int);
					break;
				case -1:
					// u = va_arg(ap, unsigned short);
					u = va_arg(ap, int);
					break;
				case 0:
					u = va_arg(ap, unsigned int);
					break;
				case 1:
					u = va_arg(ap, unsigned long);
					break;
#ifndef	NO_LONG_LONG
				case 2:
					u = va_arg(ap, unsigned long long);
					break;
#endif
				default:
					goto default_case;
				}
				/*
				 * FALLTHROUGH 
				 */
			      common2:
				s = umaxtostr(buf_end, u, base, digits);
				len = buf_end - s;
				/*
				 * the field may overflow prec 
				 */
				if (prec < len)
					/*
					 * FALLTHOUGH 
					 */
				      common3:
					prec = len;
				if (zeropad && prec < width - prefixlen)
					prec = width - prefixlen;
				else if (alt && base == 8 && u != 0LL)
					prec++;
				{
					int             tmp;
					tmp = fprintf1(output, s, len,
						       prefix, prefixlen,
						       width, prec);
					r += tmp;
					output += tmp;
				}
				break;
			default:	/* unrecognized conversion
					 * specifier */
			      default_case:
				/*
				 * print uninterpreted 
				 */
				for (s = p - 1; *s != '%'; s--);
				for (; s <= p; s++, r++) {
					// fputc(*s, f);
					*output++ = *s;
					if (r == n)
						goto overflow;
				}
				break;
			}
			break;	/* finished the conversion specifier */
		}
	}
	*output++ = '\0';
	return r;
      overflow:
	*output++ = '\0';
	return -1;
}
