/*
 * bfrdriver.h -- generic network driver interface definitions
 *
 * Use POSIX semaphores as implemented in Linux and Kenge-on-L4
 * @LICENCE@
 *
 */
#ifndef BFRDRIVER_H_
#define BFRDRIVER_H_
#include <assert.h>
#include <driver_env_glue.h>
#include <pci_dma.h>

#include "driver_interface.h"
#include "bfrdriver_types.h"
#include "lock.h"

/* Hwbuffer stats */
struct hwbuffer_stats {
	/* Nothing yet */
};

/* Combined stats */
struct bfrdriver_stats {
	struct hwbuffer_stats *hwbuffer;
	struct netstats *netdriver;
};

/* PACKET GEN */
struct gen_stats {
	volatile unsigned long packets_transmitted;
};

#define PACK_STRUCT_FIELD(x) x __attribute__((packed))
#define PACK_STRUCT_STRUCT __attribute__((packed))

struct hwbuf_eth_addr {
  PACK_STRUCT_FIELD(uint8_t addr[6]);
} PACK_STRUCT_STRUCT;

struct hwbuf_eth_hdr {
  PACK_STRUCT_FIELD(struct hwbuf_eth_addr dest);
  PACK_STRUCT_FIELD(struct hwbuf_eth_addr src);
  PACK_STRUCT_FIELD(uint16_t type);
} PACK_STRUCT_STRUCT;

struct packet_gen {
	int enabled;
	unsigned short packet_size;
	struct hwbuf_eth_hdr gen_header;
	struct gen_stats *stats;
};
/* End packet gen */

struct bfr_drv;

struct buffer_ops {
	/* Initialisation */
	struct bfr_drv *(*bfr_init)(dma_handle_t dma);
	void (*cleanup)(struct bfr_drv *bufman);

	/* Driver interface */
	int (*register_if)(struct bfr_drv *buf_mgr, void *etherif,
			uint32_t num_recv_buffers,
			uint32_t recv_buffer_size,
			uint32_t num_tx_buffers);
	int (*freepacket)(struct bfr_drv *buf_mgr, dma_addr_t addr, uintptr_t key);
	void (*feedpacket)(struct bfr_drv *buf_mgr, dma_addr_t buffer_addr, size_t size, uintptr_t internal_key);
	void (*freepacket_batch_done)(struct bfr_drv *buf_mgr);
	void (*feedpacket_batch_done)(struct bfr_drv *buf_mgr);

	/* Client side Interface */
	int (*register_client)(struct bfr_drv *buf_mgr, 
			       void *client_handle, 
			       void (*input)(void *client_handle, 
					     mem_handle_t handle,
					     size_t len, uintptr_t key),
			 	void (*transmitted)(void *client_handle, 
						    callback_data_t data,
						    uintptr_t client_key) );

	int (*rx_freepacket)(struct bfr_drv *buf_mgr, uintptr_t key);
	int (*transmit_packet)(struct bfr_drv *bufman, callback_data_t notify, packet_t *packet);
	int (*transmit_pkt_seq)(struct bfr_drv *bufman, callback_data_t notify,pkt_seq *packet);
	
	/* Packet generator */
	int (*packet_gen_start)(struct bfr_drv *buf_mgr, uint8_t *target_mac,
		unsigned short packet_size);
	void (*packet_gen_stop)(struct bfr_drv *buf_mgr);

	/* Statistics */
	void (*collect_stats)(struct bfr_drv* buf_mgr, void *statbuf);
	
	/* Tweaks */
	void (*set_discard)(struct bfr_drv *buf_mgr, int discard);
	void (*set_echo)(struct bfr_drv *buf_mgr, int echo);

	int (*getmac)(struct bfr_drv *bufman, unsigned char *buf);
};

static inline void
bfr_cleanup(struct bfr_drv *bp)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	dp->classp->bfr_ops->cleanup(bp);
}

static inline
int bfr_register_if(struct bfr_drv *bp, void *etherif, uint32_t num_recv_buffers,
	uint32_t recv_buffer_size, uint32_t num_tx_buffers)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->register_if(bp, etherif,
						num_recv_buffers,
						recv_buffer_size,
						num_tx_buffers);
}

static inline int
bfr_freepacket(struct bfr_drv *bp, dma_addr_t addr, uintptr_t key)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->freepacket(bp, addr, key);
}

static inline void
bfr_feedpacket(struct bfr_drv *bp, dma_addr_t buffer_addr, size_t sz, uintptr_t key)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	dp->classp->bfr_ops->feedpacket(bp, buffer_addr, sz, key);
}


static inline int
bfr_register_client(struct bfr_drv *bp, void *client_handle,
	void (*input)(void *client_handle, mem_handle_t handle, size_t len, uintptr_t key),
	void (*transmitted)(void *client_handle, callback_data_t data, uintptr_t client_key) )
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->register_client
		(bp, client_handle, input, transmitted);
}

static inline int
bfr_rx_freepacket(struct bfr_drv *bp,
	uintptr_t key)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->rx_freepacket(bp, key);
}

static inline int
bfr_transmit_packet(struct bfr_drv *bp,
	callback_data_t notify, packet_t *packet)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->transmit_packet(bp, notify, packet);
}

static inline int
bfr_transmit_pkt_seq(struct bfr_drv *bp,
	callback_data_t notify,	pkt_seq *pk_seq)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->transmit_pkt_seq(bp, notify, pk_seq);
}

static inline int
bfr_getmac(struct bfr_drv *bp, unsigned char *buf)
{
	struct driver_instance *dp = (struct driver_instance *)bp;
	return dp->classp->bfr_ops->getmac(bp, buf);
}

static inline int
bfr_packet_gen_start(struct bfr_drv *buf_mgr, uint8_t *target_mac,
	unsigned short packet_size)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	return dp->classp->bfr_ops->packet_gen_start(buf_mgr, target_mac, packet_size);
}

static inline void
bfr_packet_gen_stop(struct bfr_drv *buf_mgr)
{	
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	return dp->classp->bfr_ops->packet_gen_stop(buf_mgr);
}

static inline void
bfr_collect_stats(void * buf_mgr, void *statbuf)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	if (dp->classp->bfr_ops->collect_stats) {
		return dp->classp->bfr_ops->collect_stats(buf_mgr, statbuf);
	}
}

static inline void
bfr_set_discard(void *buf_mgr, int discard)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	dp->classp->bfr_ops->set_discard(buf_mgr, discard);
}

static inline void
bfr_set_echo(void *buf_mgr, int echo)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	dp->classp->bfr_ops->set_echo(buf_mgr, echo);
}

static inline void
bfr_feedpacket_batch_done(void *buf_mgr)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	if (dp->classp->bfr_ops->feedpacket_batch_done != NULL) {
		dp->classp->bfr_ops->feedpacket_batch_done(buf_mgr);
	}
}

static inline void
bfr_freepacket_batch_done(void *buf_mgr)
{
	struct driver_instance *dp = (struct driver_instance *)buf_mgr;
	if (dp->classp->bfr_ops->freepacket_batch_done != NULL) {
		dp->classp->bfr_ops->freepacket_batch_done(buf_mgr);
	}
}

int setup_dma_pktseq( dma_handle_t dma, pkt_seq *pktseq, 
		      dmabuf_handle_t *dmabuffers, direction_t dir );

/* FIXME: this garbage broke mungi. - Chuck */
/*
static inline int
setup_dma_pktseq(dma_handle_t dma, pkt_seq *pktseq, dmabuf_handle_t *dmabuffers, direction_t dir)
{
	return -1;
}
*/

#endif /* BFRDRIVER_H_ */
