#include "driver_env_glue.h"
#include <linux/pci.h>
#include <linux/mm.h> /* for mem_map required by virt_to_phys */
#include "pci_dma.h"
#include "assert.h"


callback_data_t null_callback_data = NULL;
int pagesize = PAGE_SIZE;


mem_handle_t new_mem_handle(void *addr, size_t sz) 
{
	mem_handle_t handle;
	unsigned offset = (unsigned long) addr & ~PAGE_MASK;
	unsigned i;

	/*
	 * Make a scatter-gather list with one struct for each page.
	 * This gets a little fiddly, to deal with unaligned addr.
	 */
	handle.nents = ((sz + offset) + PAGE_SIZE - 1) >> PAGE_SHIFT;
	if (!sz) {
		handle.list = handle.origlist = NULL;
		handle.size = handle.nents = 0;
		return handle;
	}
	handle.list = handle.origlist = kmalloc(
		sizeof(struct scatterlist) * handle.nents, GFP_KERNEL);
	if (handle.list == NULL) {
		handle.nents = 0;
		handle.size = handle.nents = 0;
		printk("malloc failed!\n");
		return handle;
	}
	handle.list[0].page = virt_to_page(addr);
	handle.list[0].offset = offset;
	handle.list[0].length = PAGE_SIZE - offset;

	for (i = 1; i < handle.nents; ++i) {
		addr += PAGE_SIZE;
		handle.list[i].page = virt_to_page(addr);
		handle.list[i].length = PAGE_SIZE;
	}
	handle.list[handle.nents-1].length 
		= ((sz + offset - 1) & ~PAGE_MASK) + 1;
	handle.list[0].length -= offset;
	handle.size = sz;
	
	return handle;
}

void free_mem_handle(mem_handle_t *handle)
{
	if (handle->origlist != NULL)
		kfree(handle->origlist);
}

mem_handle_t handle_increment(mem_handle_t *handle, size_t sz)
{
	handle->size -= sz;
	if (handle->size < 0)
	    handle->size = 0;
	
	/* moves ->list forwards, and adjust ->nents accordingly */
	while (handle->list != NULL) {
		if (sz < handle->list->length) {
			handle->list->offset += sz;
			break;
		} else if (handle->nents > 1) {
			sz -= handle->list->length;
			++handle->list;
			--handle->nents;
		} else
			handle->list = NULL;
	}
	return *handle;
}

int get_pinnable_memory(size_t size, mem_handle_t *handle)
{
	/* this function could be improved, but it's not used much */
	unsigned i;
	char * mem;
	handle->nents = (size + PAGE_SIZE-1) >> PAGE_SHIFT;
	handle->list = handle->origlist = kmalloc(
		sizeof(struct scatterlist) * handle->nents, GFP_KERNEL);
	if (handle->list == NULL) {
		handle->nents = 0;
		return 1;
	}
	//mem = vmalloc_32(size);
	mem = kmalloc(size, GFP_KERNEL);
	assert(((unsigned long)mem & ~PAGE_MASK) == 0);
	if (mem == NULL) {
		handle->nents = 0;
		kfree(handle->list);
		return 1;
	}
	for (i = 0; i < handle->nents; ++i) {
		handle->list[i].page = virt_to_page(mem + (i << PAGE_SHIFT));
		handle->list[i].offset = 0;
		handle->list[i].length = PAGE_SIZE;
	}
	handle->list[handle->nents-1].length = ((size-1) & ~PAGE_MASK) + 1;
	handle->size = size;
	return 0;
}

int setup_dma(dma_handle_t dma,
              mem_handle_t handle,
              size_t size,
	      dmabuf_handle_t *dmabuf,
              direction_t dir)
{
	int sg_nents, i;

	/* make dmabuf a prefix of handle */
	dmabuf->list = handle.list;
	if (size < handle.size) {
		size_t dmabuf_size;
		dmabuf_size = dmabuf->nents = 0;
		while (dmabuf_size < size) {
			assert(dmabuf->nents < handle.nents);
			dmabuf_size += handle.list[dmabuf->nents++].length;
		}
	} else
		dmabuf->nents = handle.nents;

	/* convert from benno's direction to linux direction constants */
	switch (dir) {
	case mem_to_bus: dmabuf->direction = PCI_DMA_TODEVICE;      break;
	case bus_to_mem: dmabuf->direction = PCI_DMA_FROMDEVICE;    break;
	case consistent: dmabuf->direction = PCI_DMA_BIDIRECTIONAL; break;
	default:         dmabuf->direction = PCI_DMA_NONE;          break;
	}

	/* obtain dma mapping from linux */
	sg_nents = pci_map_sg(dma, dmabuf->list, dmabuf->nents,
		dmabuf->direction);

	/* On some architectures. sg_dma_len is different from length field,
	 * so copy it to length (so that handle_increment still works).     */
	for (i = 0; i < sg_nents; ++i)
		dmabuf->list[i].length = sg_dma_len(&dmabuf->list[i]);

	/* zero out any trailing entries so handle_increment still works */
	for (/* i = sg_nents */; i < dmabuf->nents; ++i)
		dmabuf->list[i].length = 0;
	
	dmabuf->nents = sg_nents;
	
	return 0;
}

int teardown_dma(dma_handle_t dma, dmabuf_handle_t *dmabuf)
{
	/* tell linux that we're finished with this mapping */
	pci_unmap_sg(dma, dmabuf->list, dmabuf->nents, dmabuf->direction);
	return 0;
}

int get_pci_consistent_memory(dma_handle_t dma, size_t sz, mem_handle_t *vaddr,
	dmabuf_handle_t *dmabuf)
{
	int status;
	status = get_pinnable_memory(sz, vaddr);
	if (!status)
		status = setup_dma(dma, *vaddr, sz, dmabuf, consistent);
	return status;
}
