#include "LinuxTimes.h"
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#define MAXTHREADS 100

// number of threads being monitored
static int m_nthreads;

// fd open to /proc/pid/stat for each thread
static int m_proc_pid_stat_fd[MAXTHREADS];

// make monitorme threadsafe
static pthread_mutex_t m_lock;



/**
 *   New threads call this to add themselves to the group.
 *   Threadsafe.
 *   @return 0 on success, Unix error code on failure
 */
int monitorme(void) 
{
	int fd;
	char fname[256];
	int err = 0;

	if (pthread_mutex_lock(&m_lock))
		return EINVAL;

	if (m_nthreads >= MAXTHREADS) {
		err = E2BIG;
	} else {
		// Under LinuxThreads, each thread has its own pid
		sprintf(fname, "/proc/%d/stat", getpid());
		fd = open(fname, O_RDONLY);
		if (fd == -1)
			err = errno;
		else {
			m_proc_pid_stat_fd[m_nthreads++] = fd;
		}
	}

	if (pthread_mutex_unlock(&m_lock))
		return EINVAL;

	return err;
}

/* Skip to char after nth whitespace. Returns NULL on failure. */
static inline const char *skipNspace(const char *p, int n)
{
	while (*++p)
		if (isspace(*p) && ! --n)
			return p+1;
	return NULL;
}

/**
 *   Calculate user and system time accumulated by all threads in
 *   group.  Return result in tms_utime and tms_stime fields of given
 *   struct tms.  Similar to 'man 2 times' on Solaris (where all CPU
 *   time of all threads * is counted as CPU time towards the
 *   process).   @return 0 on success, Unix error code on failure
*/
int
LinuxTimes(struct tms *buf)
{
	int i;
	int nread;

	buf->tms_utime = 0;
	buf->tms_cutime = 0;
	buf->tms_stime = 0;
	buf->tms_cstime = 0;
	for (i=0; i < m_nthreads; i++) {
		char scratch[1024]; // FIXME: is that big enough?
		int fd = m_proc_pid_stat_fd[i];

		// rewind to start of stat file. (Not all /proc entries support this.)
		if (lseek(fd, 0, SEEK_SET))
			return EBADF;

		// Read in ASCII data and parse out utime and stime fields
		// (see 'man proc')
		nread = read(fd, scratch, sizeof(scratch)-1);
		if (nread < 50) // FIXME: cheesy
			return EINVAL;
		scratch[nread] = 0;

		// Skip to end of command field
		// FIXME: what if executable has ) in its filename? Bleh.
		const char *p = strchr(scratch, ')') + 2;

		// Skip to utime field
		p = skipNspace(p, 11);
		if (!p) return EINVAL;
		buf->tms_utime += atoi(p);

		//fprintf(stderr, "%d uticks, ", atoi(p));

		// Skip to stime field
		p = skipNspace(p, 1);
		if (!p) return EINVAL;
		buf->tms_stime += atoi(p);

		//fprintf(stderr, "%d sticks\n", atoi(p));

		// Skip to cutime field
		p = skipNspace(p, 1);
		if (!p) return EINVAL;
		buf->tms_cutime += atoi(p);

		// Skip to cstime field
		p = skipNspace(p, 1);
		if (!p) return EINVAL;
		buf->tms_cstime += atoi(p);
	}

	return 0;
}

void
LinuxTimes_Init(void)
{
	m_nthreads = 0; 
	pthread_mutex_init(&m_lock, NULL); 
}
