#include <pthread.h>
#include <stdint.h>
#include "error.h"
#include <unistd.h>
#include <string.h> /* strerror() */
#include <errno.h> /* errno */
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h> /* snprintf */
#include <poll.h>
#include "setrealtime.h"
#include "usr.h"

static void *interrupt_thread(void *arg);
static pthread_t ithr;
unsigned long interrupts = 0;

struct irq_desc {
	void   *driverp;
	int    (*handler)(void *);
	int	irqfd;
};


static void *
interrupt_thread(void *arg)
{
	struct irq_desc *ip = (struct irq_desc *)arg;
	int nirq;
	long err;
	int fd = ip->irqfd;

#ifdef IRQ_USE_POLL
	struct pollfd fds;
	int ret;

	fds.fd = fd;
	fds.events = POLLIN;
#endif

	/* Set realtime priority explicitly here, as
	 * NPTL is broken as yet
	 */
	set_realtime_priority(2);
	for (;;) {
#ifdef IRQ_USE_POLL
		ret = poll(&fds, 1, -1);
		if ( ret > 0 )
		{
			if (fds.revents & POLLIN)
			{
#endif
				err = read(fd, &nirq, sizeof(nirq));	     
				if (err == -1)
					switch(errno)
					{
					case EINTR:
						continue;
					default:
						syserror("Interrupt device read returned EINVAL");
						/* NOTREACHED */
						break;
					case EBUSY:
					{
						int enable = 0;
						write(fd, &enable, sizeof enable);
						continue;
					}
					}
				interrupts++;
				(void)(ip->handler)(ip->driverp);
#ifdef IRQ_USE_POLL
			}
		}
#endif
	}
	/* NOTREACHED */
	return NULL;
}	

int
create_interrupt_thread(int irq, void *drive, int (*handler)(void *))
{
	int fd;
	char name[24];
	struct irq_desc *idp = malloc(sizeof(*idp));
	pthread_attr_t attr;
	struct sched_param schp;

	assert(ithr == 0);
	assert(idp != NULL);
	snprintf(name, sizeof name,  "/proc/irq/%d/irq", irq);
	if ((fd = open(name, O_RDWR|O_EXCL)) == -1)
		syserror("Cannot open interrupt descriptor");
	idp->irqfd = fd;
	idp->driverp = drive;
	idp->handler = handler;

	memset(&schp, 0, sizeof(schp));
	schp.sched_priority = sched_get_priority_max(SCHED_FIFO);

	pthread_attr_init(&attr);
	pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	pthread_attr_setschedparam(&attr, &schp);
	pthread_create(&ithr, &attr, interrupt_thread, idp);

	return 0;
}

void
destroy_interrupt_thread(void)
{
	assert(ithr != 0);
	pthread_cancel(ithr);
	pthread_join(ithr, NULL);
	ithr = 0;
}
