/******************************************************************************
 *
 * Module Name: acpiosxf - OSL interfaces
 *
 *****************************************************************************/

#include <malloc.h>
#include <sprintf.h>
#include <assert.h>
#include "driver_env_glue.h"
#include "io/io.h"

#include "acpi.h"
#include "amlcode.h"
#include "acparser.h"
#include "acdebug.h"

#define _COMPONENT          ACPI_OS_SERVICES
ACPI_MODULE_NAME    ("acpiosxf")

#define NOT_IMPLEMENTED(fn)	UserPrint("ACPI: " fn " not implemented\n")

static UINT32
pci_config_address (ACPI_PCI_ID *pci_id, UINT32 reg)
{
	return (pci_id->Segment << 24)
		| (pci_id->Bus << 16)
		| (pci_id->Device << 11)
		| (pci_id->Function << 8)
		| reg;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsInitialize, AcpiOsTerminate
 *
 * PARAMETERS:  None
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Init and terminate.  Nothing to do.
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsInitialize (void)
{
	return AE_OK;
}


ACPI_STATUS
AcpiOsTerminate (void)
{
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsGetRootPointer
 *
 * PARAMETERS:  Flags   - Logical or physical addressing mode
 *              Address - Where the address is returned
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Gets the root pointer (RSDP)
 *
 *****************************************************************************/

extern uintptr_t acpi_root;

ACPI_STATUS
AcpiOsGetRootPointer (
		      UINT32                  Flags,
		      ACPI_POINTER           *Address)
{
	Address->PointerType = ACPI_PHYSICAL_POINTER;
	Address->Pointer.Physical = acpi_root;
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsTableOverride
 *
 * PARAMETERS:  ExistingTable   - Header of current table (probably firmware)
 *              NewTable        - Where an entire new table is returned.
 *
 * RETURN:      Status, pointer to new table.  Null pointer returned if no
 *              table is available to override
 *
 * DESCRIPTION: Return a different version of a table if one is available
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsTableOverride (
		     ACPI_TABLE_HEADER       *ExistingTable,
		     ACPI_TABLE_HEADER       **NewTable)
{
	*NewTable = NULL;
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadable
 *
 * PARAMETERS:  Pointer             - Area to be verified
 *              Length              - Size of area
 *
 * RETURN:      TRUE if readable for entire length
 *
 * DESCRIPTION: Verify that a pointer is valid for reading
 *
 *****************************************************************************/

BOOLEAN
AcpiOsReadable (
		void                    *Pointer,
		UINT32                  Length)
{
	return (TRUE);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWritable
 *
 * PARAMETERS:  Pointer             - Area to be verified
 *              Length              - Size of area
 *
 * RETURN:      TRUE if writable for entire length
 *
 * DESCRIPTION: Verify that a pointer is valid for writing
 *
 *****************************************************************************/

BOOLEAN
AcpiOsWritable (
		void                    *Pointer,
		UINT32                  Length)
{
	return (TRUE);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsRedirectOutput
 *
 * PARAMETERS:  Destination         - An open file handle/pointer
 *
 * RETURN:      None
 *
 * DESCRIPTION: Causes redirect of AcpiOsPrintf and AcpiOsVprintf
 *
 *****************************************************************************/

void
AcpiOsRedirectOutput (
		      void                    *Destination)
{
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsPrintf
 *
 * PARAMETERS:  fmt, ...            Standard printf format
 *
 * RETURN:      None
 *
 * DESCRIPTION: Formatted output
 *
 *****************************************************************************/

void ACPI_INTERNAL_VAR_XFACE
AcpiOsPrintf (
	      const char              *Fmt,
	      ...)
{
	va_list                 Args;

	va_start(Args, Fmt);
	AcpiOsVprintf(Fmt, Args);
	va_end(Args);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsVprintf
 *
 * PARAMETERS:  fmt                 Standard printf format
 *              args                Argument list
 *
 * RETURN:      None
 *
 * DESCRIPTION: Formatted output with argument list pointer
 *
 *****************************************************************************/

void
AcpiOsVprintf (
	       const char              *Fmt,
	       va_list                 Args)
{
	char buffer[256];

	if (vsnprintf(buffer, sizeof(buffer), Fmt, Args) > 0)
		UserPrint("%s", buffer);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsGetLine
 *
 * PARAMETERS:  fmt                 Standard printf format
 *              args                Argument list
 *
 * RETURN:      Actual bytes read
 *
 * DESCRIPTION: Formatted input with argument list pointer
 *
 *****************************************************************************/

UINT32
AcpiOsGetLine (
	       char                    *Buffer)
{
#if 0
	UINT32                  i;

	for (i = 0; ; i++)
	{
		Buffer[i] = getc();
		if (!Buffer[i] || Buffer[i] == '\n')
			break;
	}

	Buffer[i] = 0;
	return (i);
#endif
	return 0;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsMapMemory
 *
 * PARAMETERS:  where               Physical address of memory to be mapped
 *              length              How much memory to map
 *              there               Logical address of mapped memory
 *
 * RETURN:      Pointer to mapped memory.  Null on error.
 *
 * DESCRIPTION: Map physical memory into caller's address space
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsMapMemory (
		 ACPI_PHYSICAL_ADDRESS   where,
		 ACPI_SIZE               length,
		 void                    **there)
{
	*there = (void *)where;
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsUnmapMemory
 *
 * PARAMETERS:  where               Logical address of memory to be unmapped
 *              length              How much memory to unmap
 *
 * RETURN:      None.
 *
 * DESCRIPTION: Delete a previously created mapping.  Where and Length must
 *              correspond to a previous mapping exactly.
 *
 *****************************************************************************/

void
AcpiOsUnmapMemory (
		   void                    *where,
		   ACPI_SIZE               length)
{
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsAllocate
 *
 * PARAMETERS:  Size                Amount to allocate, in bytes
 *
 * RETURN:      Pointer to the new allocation.  Null on error.
 *
 * DESCRIPTION: Allocate memory.  Algorithm is dependent on the OS.
 *
 *****************************************************************************/

void *
AcpiOsAllocate (
		ACPI_SIZE               size)
{
	return malloc(size);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsFree
 *
 * PARAMETERS:  mem                 Pointer to previously allocated memory
 *
 * RETURN:      None.
 *
 * DESCRIPTION: Free memory allocated via AcpiOsAllocate
 *
 *****************************************************************************/

void
AcpiOsFree (
	    void                    *mem)
{
	free(mem);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsCreateSemaphore
 *
 * PARAMETERS:  InitialUnits        - Units to be assigned to the new semaphore
 *              OutHandle           - Where a handle will be returned
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Create an OS semaphore
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsCreateSemaphore (
		       UINT32              MaxUnits,
		       UINT32              InitialUnits,
		       ACPI_HANDLE         *OutHandle)
{
	int *semaphore = AcpiOsAllocate(sizeof(int));
	*semaphore = InitialUnits;
	*OutHandle = (ACPI_HANDLE)semaphore;
	return AE_OK;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsDeleteSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Delete an OS semaphore
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsDeleteSemaphore (
		       ACPI_HANDLE         Handle)
{
	AcpiOsFree((void *)Handle);
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWaitSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *              Units               - How many units to wait for
 *              Timeout             - How long to wait
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Wait for units
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsWaitSemaphore (
		     ACPI_HANDLE         Handle,
		     UINT32              Units,
		     UINT16              Timeout)
{
	int *semaphore = (int *)Handle;

	/* HACK */
	while (1)
	{
		if (*semaphore >= Units)
		{
			*semaphore -= Units;
			return AE_OK;
		}
		if (Timeout == 0)
			return AE_TIME;
		AcpiOsSleep(0, 1);
		Timeout--;
	}
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsSignalSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *              Units               - Number of units to send
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Send units
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsSignalSemaphore (
		       ACPI_HANDLE         Handle,
		       UINT32              Units)
{
	int *semaphore = (int *)Handle;
	*semaphore += Units;
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsInstallInterruptHandler
 *
 * PARAMETERS:  InterruptNumber     Level handler should respond to.
 *              Isr                 Address of the ACPI interrupt handler
 *              ExceptPtr           Where status is returned
 *
 * RETURN:      Handle to the newly installed handler.
 *
 * DESCRIPTION: Install an interrupt handler.  Used to install the ACPI
 *              OS-independent handler.
 *
 *****************************************************************************/

UINT32
AcpiOsInstallInterruptHandler (
			       UINT32                  InterruptNumber,
			       OSD_HANDLER             ServiceRoutine,
			       void                    *Context)
{
	UserPrint("ACPI: subsystem requested IRQ %u\n", InterruptNumber);
	NOT_IMPLEMENTED("AcpiOsInstallInterruptHandler");
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsRemoveInterruptHandler
 *
 * PARAMETERS:  Handle              Returned when handler was installed
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Uninstalls an interrupt handler.
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsRemoveInterruptHandler (
			      UINT32                  InterruptNumber,
			      OSD_HANDLER             ServiceRoutine)
{
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsQueueForExecution
 *
 * PARAMETERS:  Priority        - Requested execution priority
 *              Function        - Address of the function to execute
 *              Context         - Passed as a parameter to the function
 *
 * RETURN:      Status.
 *
 * DESCRIPTION: Sleep at microsecond granularity
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsQueueForExecution (
			 UINT32                  Priority,
			 OSD_EXECUTION_CALLBACK  Function,
			 void                    *Context)
{
	NOT_IMPLEMENTED("AcpiOsQueueForExecution");
	return 0;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsBreakpoint
 *
 * PARAMETERS:  Msg                 Message to print
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Print a message and break to the debugger.
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsBreakpoint (char *Msg);

ACPI_STATUS
AcpiOsBreakpoint (
		  char                    *Msg)
{
	AcpiOsPrintf ("ACPI: Breakpoint: %s\n", (Msg != NULL) ? Msg : "");
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsSleepUsec
 *
 * PARAMETERS:  microseconds        To sleep
 *
 * RETURN:      Blocks until sleep is completed.
 *
 * DESCRIPTION: Sleep at microsecond granularity
 *
 *****************************************************************************/

void
AcpiOsStall (
	     UINT32                  microseconds)
{
	UINT64 i;
	for (i = 0; i < (UINT64)microseconds * 100000; i++);

	/* On some machines such as HP i2000, ACPI initialisation
	 * invokes firmware via a PMI.  However, since vNUMA normally
	 * runs with ic off, PMIs are not delivered asynchronously.
	 * Instead, sample_pmi is used to deliver PMIs here.
	 */
	sample_pmi();
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsSleep
 *
 * PARAMETERS:  seconds             To sleep
 *              milliseconds        To sleep
 *
 * RETURN:      Blocks until sleep is completed.
 *
 * DESCRIPTION: Sleep at second/millisecond granularity
 *
 *****************************************************************************/

void
AcpiOsSleep (
	     UINT32                  seconds,
	     UINT32                  milliseconds)
{
	AcpiOsStall( seconds * 1000000 + milliseconds * 1000);
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsGetTimer
 *
 * PARAMETERS:  None
 *
 * RETURN:      Current time in milliseconds
 *
 * DESCRIPTION: Get the current system time (in milliseconds).
 *
 *****************************************************************************/

UINT32
AcpiOsGetTimer (void)
{
	NOT_IMPLEMENTED("AcpiOsGetTimer");
	return 0;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadPciConfiguration
 *
 * PARAMETERS:  PciId               Seg/Bus/Dev
 *              Register            Device Register
 *              Value               Buffer where value is placed
 *              Width               Number of bits
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Read data from PCI configuration space
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsReadPciConfiguration (
			    ACPI_PCI_ID             *PciId,
			    UINT32                  Register,
			    void                    *Value,
			    UINT32                  Width)
{
	UINT32 Address = pci_config_address(PciId, Register);
	switch (Width)
	{
		case 8:
			*((UINT8 *) Value) = pci_read_config_byte(Address);
			break;

		case 16:
			*((UINT16 *) Value) = pci_read_config_word(Address);
			break;

		case 32:
			*((UINT32 *) Value) = pci_read_config_dword(Address);
			break;

		default:
			printf("AcpiOsReadPciConfiguration: Width = %u\n", Width);
	}
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWritePciConfiguration
 *
 * PARAMETERS:  PciId               Seg/Bus/Dev
 *              Register            Device Register
 *              Value               Value to be written
 *              Width               Number of bits
 *
 * RETURN:      Status.
 *
 * DESCRIPTION: Write data to PCI configuration space
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsWritePciConfiguration (
			     ACPI_PCI_ID             *PciId,
			     UINT32                  Register,
			     ACPI_INTEGER            Value,
			     UINT32                  Width)
{
	UINT32 Address = pci_config_address(PciId, Register);
	switch (Width)
	{
		case 8:
			pci_write_config_byte(Address, Value);
			break;

		case 16:
			pci_write_config_word(Address, Value);
			break;

		case 32:
			pci_write_config_dword(Address, Value);
			break;

		default:
			printf("AcpiOsWritePciConfiguration: Width = %u\n", Width);
	}
	return AE_OK;
}

/* TEMPORARY STUB FUNCTION */
void
AcpiOsDerivePciId(
		  ACPI_HANDLE	            rhandle,
		  ACPI_HANDLE             chandle,
		  ACPI_PCI_ID	            **PciId)
{
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadPort
 *
 * PARAMETERS:  Address             Address of I/O port/register to read
 *              Value               Where value is placed
 *              Width               Number of bits
 *
 * RETURN:      Value read from port
 *
 * DESCRIPTION: Read data from an I/O port or register
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsReadPort (
		ACPI_IO_ADDRESS         Address,
		void                    *Value,
		UINT32                  Width)
{
	switch (Width)
	{
		case 8:
			*((UINT8 *) Value) = inb(Address);
			break;

		case 16:
			*((UINT16 *) Value) = inw(Address);
			break;

		case 32:
			*((UINT32 *) Value) = inl(Address);
			break;

		default:
			printf("AcpiOsReadPort: Width = %u\n", Width);
	}
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWritePort
 *
 * PARAMETERS:  Address             Address of I/O port/register to write
 *              Value               Value to write
 *              Width               Number of bits
 *
 * RETURN:      None
 *
 * DedyESCRIPTION: Write data to an I/O port or register
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsWritePort (
		 ACPI_IO_ADDRESS         Address,
		 ACPI_INTEGER            Value,
		 UINT32                  Width)
{
	switch (Width)
	{
		case 8:
			outb(Address, Value);
			break;

		case 16:
			outw(Address, Value);
			break;

		case 32:
			outl(Address, Value);
			break;

		default:
			printf("AcpiOsWritePort: Width = %u\n", Width);
	}
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadMemory
 *
 * PARAMETERS:  Address             Physical Memory Address to read
 *              Value               Where value is placed
 *              Width               Number of bits
 *
 * RETURN:      Value read from physical memory address
 *
 * DESCRIPTION: Read data from a physical memory address
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsReadMemory (
		  ACPI_PHYSICAL_ADDRESS   Address,
		  void                    *Value,
		  UINT32                  Width)
{
	switch (Width)
	{
		case 8:
			*((UINT8 *) Value) = readb(Address);
			break;

		case 16:
			*((UINT16 *) Value) = readw(Address);
			break;

		case 32:
			*((UINT32 *) Value) = readl(Address);
			break;

		default:
			printf("AcpiOsReadMemory: Width = %u\n", Width);
	}
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWriteMemory
 *
 * PARAMETERS:  Address             Physical Memory Address to write
 *              Value               Value to write
 *              Width               Number of bits
 *
 * RETURN:      None
 *
 * DESCRIPTION: Write data to a physical memory address
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsWriteMemory (
		   ACPI_PHYSICAL_ADDRESS   Address,
		   ACPI_INTEGER            Value,
		   UINT32                  Width)
{
	switch (Width)
	{
		case 8:
			writeb(Address, Value);
			break;

		case 16:
			writew(Address, Value);
			break;

		case 32:
			writel(Address, Value);
			break;

		default:
			printf("AcpiOsWriteMemory: Width = %u\n", Width);
	}
	return AE_OK;
}


UINT32
AcpiOsGetThreadId(void)
{
	return 0;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsSignal
 *
 * PARAMETERS:  Function            ACPI CA signal function code
 *              Info                Pointer to function-dependent structure
 *
 * RETURN:      Status
 *
 * DESCRIPTION: Miscellaneous functions
 *
 *****************************************************************************/

ACPI_STATUS
AcpiOsSignal (
	      UINT32                  Function,
	      void                    *Info)
{
	switch (Function)
	{
		case ACPI_SIGNAL_FATAL:
			AcpiOsBreakpoint("Fatal opcode");
			break;

		case ACPI_SIGNAL_BREAKPOINT:
		default:
			AcpiOsBreakpoint(Info);
			break;
	}

	return AE_OK;
}


