#include "vm.h"
#include <stdarg.h>
#include <sprintf.h>
#include <driver_interface.h>
#include <uart16550_interface.h>

extern void *serial, *ether_drv;
extern unsigned int node_id;
extern char *last_interruption;

#define LOG_BUFFER_SIZE 16384
char log_buffer[LOG_BUFFER_SIZE];
unsigned int log_buffer_next;

void putchar(int c)
{
	uart16550_putchar(serial, c);
}

int puts(const char *s)
{
	while (*s != 0)
	{
		putchar(*s);
		s++;
	}
	return 0;
}

int printf(const char *format, ...)
{
	char buffer[1024];
	va_list ap;
	int rv;

	va_start(ap, format);
	rv = vsnprintf(buffer, sizeof(buffer), format, ap);
	puts(buffer);
	va_end(ap);
	return rv;
}

int logf(const char *format, ...)
{
	char buffer[1024];
	va_list ap;
	unsigned int len, rv;

	va_start(ap, format);
	rv = vsnprintf(buffer, sizeof(buffer), format, ap);

	len = LOG_BUFFER_SIZE-log_buffer_next; 
	if (rv > len)
	{
		memcpy(&log_buffer[log_buffer_next], buffer, len);
		memcpy(log_buffer, buffer+len, rv-len);
		log_buffer_next = rv-len;
	}
	else
	{
		memcpy(&log_buffer[log_buffer_next], buffer, rv);
		log_buffer_next += rv;
	}

	va_end(ap);
	return rv;
}

#if 0
void
log(unsigned long val)
{
	log_buffer[log_buffer_next] = val;
	log_buffer_next = (log_buffer_next + 1) % LOG_BUFFER_SIZE;
}
#endif

extern unsigned long read_gr(unsigned int regno);

static unsigned long
read_br(unsigned int regno)
{
	switch (regno)
	{
		case 1: return read_register("b1");
		case 2: return read_register("b2");
		case 3: return read_register("b3");
		case 4: return read_register("b4");
		case 5: return read_register("b5");
		case 6: return read_register("b6");
		default: printf("read_br bad register\n"); return 0;
	}
}
int process_debug_char(char c)
{
	int i;

	switch (c)
	{
	case 'r':
		ipc_reset_system();
		reset_system();
		break;
	case 'g':
		printf("go\n");
		return 1;
	case 'd':
		printf("dump state\n");
		printf("iip: %016lx\n", read_iip());
		printf("ipsr:%016lx\n", read_ipsr());
		printf("vpsr:%016lx\n", cpu.psr);
		printf("ifs: %016lx\n", read_ifs());
		printf("bsp: %016lx\n", read_bsp());
		for (i = 0; i < 8; i++)
			printf("rr%d: %016lx\n", i, read_rr((unsigned long)i << 61));
		for (i = 0; i < 8; i++)
			printf(" k%d: %016lx\n", i, cpu.kr[i]);
		for (i = 1; i < 6; i++)
			printf(" b%d: %016lx\n", i, read_br(i));
		for (i = 1; i < 40; i++)
			printf("r%02d: %016lx\n", i, read_gr(i));
		break;
#ifndef TIGON3
	case 'n':
		dp83820_reap_tx_descs(ether_drv);
		if (dp83820_rx_poll(ether_drv))
			printf("poll found packets!\n");
		break;
#endif
	case 't':
		printf("IPC_PROBE\n");
		ipc_send_reply(63, IPC_PROBE);
		break;
	case 'p':
		{
		unsigned long comm = cpu.kr[6]+724;
		if (tpa(comm) != 0)
			printf("current process: %16s\n", (char *)comm);
		else
			printf("current process indeterminate\n");
		break;
		}
	case 'b':
		printf("enabling breakpoint\n");
#if 0
		write_ibr(0, 0xa000000000010640UL);
		write_ibr(1, 0x8fffffffffffffffUL);
#endif
		write_dbr(0, 0x1580);
		write_dbr(1, 0x4fffffffffffffffUL);
		break;
	case 's':
		debug_flags ^= DEBUG_FLAG_PRINT_SHORT;
		break;
	case 'e':
		debug_flags ^= DEBUG_FLAG_PRINT_LONG;
		break;
	case 'f':
		printf("flushing\n");
		emu_fwb();
		break;
	case '1':
#ifndef TIGON3
		dp83820_dump_stats(ether_drv);
#endif
		break;
	case 'w':
		mm_dump_write_log_stats();
		break;
	case 'i':
		printf("%s\n", last_interruption);
		break;
	case 'l':
		for (i = log_buffer_next; i < log_buffer_next+LOG_BUFFER_SIZE; i++)
			putchar(log_buffer[i % LOG_BUFFER_SIZE]);
	}

	return 0;
}

void debug(void)
{
	char c;

	do {
		printf("debugger: ");
		do {
			c = uart16550_getchar_sync(serial);
		} while (!c);
	} while (!process_debug_char(c));
}

void handle_debug(void)
{
	printf("breakpoint hit: %lx %lx\n", read_iip(), read_register("r5"));
	write_ipsr(read_ipsr() | PSR_ID | PSR_DD);
}

void __assert(const char *file, int line, const char *str)
{
	printf("assertion failed: %s(%d): %s\n", file, line, str);
	printf("-----------------------------------------------");
	//debug();
}

