#include "vm.h"
#include "driver_env_glue.h"
#include "pci_dma.h"
#include "io/io.h"
#include "efi.h"
#include "sal.h"
#include "acpi.h"
#include "hcdp.h"
#include "iosapic.h"

#include <block_driver.h>
#include <netdriver.h>
#include <bfrdriver.h>
#include <mpt_interface.h>
#include <uart16550_interface.h>

void *netif, *disk, *bfr_drv, *scsi, *serial;
int pagesize = PAGESIZE;
unsigned long ia64_io_base;
unsigned int serial_gsi;
extern unsigned int node_id;

struct device {
	uint32_t pciid;
	struct driver * (*driver_init)(void);
	void (*class_setup)(struct device *device, struct driver *driver, bs_seq spaces, bus_space_t pciconf, unsigned int gsi);
	char *name;
};

struct interrupt {
	void (*handler)(void *);
	void *handler_arg;
	uintptr_t iosapic_addr;
	unsigned int rte;
	unsigned int eoi;
} interrupts[256];

static void
register_irq(unsigned int gsi, void (*handler)(void *), void *handler_arg, char *driver)
{
	struct interrupt *intr;
	unsigned int irq;
	uint32_t mode;

	if ((gsi < 16) /* ISA */ || (gsi == 0x22) /* FIXME serial on zx6000 */ || (gsi == 0x2d) /* FIXME serial on zx2000 */)
	{
		mode = (IOSAPIC_EDGE << IOSAPIC_TRIGGER_SHIFT) | (IOSAPIC_POL_HIGH << IOSAPIC_POLARITY_SHIFT) | (IOSAPIC_LOWEST_PRIORITY << IOSAPIC_DELIVERY_SHIFT);
	}
	else
	{
		mode = (IOSAPIC_LEVEL << IOSAPIC_TRIGGER_SHIFT) | (IOSAPIC_POL_LOW << IOSAPIC_POLARITY_SHIFT) | (IOSAPIC_LOWEST_PRIORITY << IOSAPIC_DELIVERY_SHIFT);
	}
	irq = gsi + 16;
	assert(irq < 0xff);

	intr = &interrupts[irq];
	intr->handler = handler;
	intr->handler_arg = handler_arg;
	assert(intr->iosapic_addr);
	writel(intr->iosapic_addr + IOSAPIC_REG_SELECT, IOSAPIC_RTE_HIGH(intr->rte));
	writel(intr->iosapic_addr + IOSAPIC_WINDOW, 0); /* cpu id */
	writel(intr->iosapic_addr + IOSAPIC_REG_SELECT, IOSAPIC_RTE_LOW(intr->rte));
	writel(intr->iosapic_addr + IOSAPIC_WINDOW, mode | irq);

	printf("PCI: Routed GSI %x to IRQ %x for %s\n", gsi, irq, driver);
}

int virtual_mode;

void
do_irq(unsigned int irq)
{
	assert(irq <= 0xff);
	if (interrupts[irq].handler)
		interrupts[irq].handler(interrupts[irq].handler_arg);
	else
		printf("spurious interrupt %d\n", irq);

	if (interrupts[irq].eoi)
	{
		if (virtual_mode)
			writel_physical(interrupts[irq].iosapic_addr + IOSAPIC_EOI, irq);
		else
			writel(interrupts[irq].iosapic_addr + IOSAPIC_EOI, irq);
	}

	write_eoi();
}

void
handle_interrupt(void)
{
	unsigned long irq;
	unsigned long start, end;

	start = read_itc();
	while (1)
	{
		irq = read_ivr();
		if (irq == 0xf)
			break;

		/* printf("irq %d\n", irq); */
		assert(irq <= 0xff);
		if (interrupts[irq].handler)
			interrupts[irq].handler(interrupts[irq].handler_arg);
		else
			printf("spurious interrupt %d\n", irq);

		if (interrupts[irq].eoi)
			writel_physical(interrupts[irq].iosapic_addr + IOSAPIC_EOI, irq);

		write_eoi();
	}

	check_interrupts();
	end = read_itc();
	CORE_STATS_ADD(interrupt_time, end-start);
	STATS_INC(interrupts);
}

static void
pci_net_setup(struct device *device, struct driver *driver, bs_seq spaces, bus_space_t pciconf, unsigned int gsi)
{
	struct driver *buffer_driver;

	if (netif != NULL)
	{
		printf("netif_setup: %s: already have a device of this type\n", device->name);
		return;
	}

	buffer_driver = ipc_init();
	bfr_drv = buffer_driver->bfr_ops->bfr_init(NULL);
	netif = driver->n_ops->setup(spaces, NULL, pciconf);
	register_irq(gsi, netif_interrupt, netif, device->name);
	/* netif_setstatistics(netif, &netstats); */
	netif_enable(netif, bfr_drv);
	netif_enable_receive(netif);
}

static void
pci_disk_setup(struct device *device, struct driver *driver, bs_seq spaces, bus_space_t pciconf, unsigned int gsi)
{
	if (disk != NULL)
	{
		printf("disk_setup: %s: already have a device of this type\n", device->name);
		return;
	}

	if (spaces.v[0].address == 0)
	{
		spaces.v[0].type = IORESOURCE_IO;
		spaces.v[0].address = 0x1f0;
		spaces.v[1].type = IORESOURCE_IO;
		spaces.v[1].address = 0x3f4;
		spaces.v[2].type = IORESOURCE_IO;
		spaces.v[2].address = 0x170;
		spaces.v[3].type = IORESOURCE_IO;
		spaces.v[3].address = 0x374;
		gsi = 14;
	}
	disk = driver->b_ops->setup(spaces, NULL, pciconf);
	register_irq(gsi, blkdev_interrupt, disk, device->name);
	blkdev_enable(disk);
}

static void
pci_scsi_setup(struct device *device, struct driver *driver, bs_seq spaces, bus_space_t pciconf, unsigned int gsi)
{
	char cdb[10];
	int ret;

	if (scsi != NULL)
	{
		printf("scsi_setup: %s: already have a device of this type\n", device->name);
		return;
	}
	scsi = mpt_setup(spaces, NULL, pciconf);
	register_irq(gsi, mpt_intr, scsi, device->name);

	printf("scsi_setup: issuing TEST_UNIT_READY...\n");
	memset(cdb, 0, 10);
	cdb[0] = 0 /* TEST_UNIT_READY */;
	ret = mpt_scsi_io_request(scsi, 0, 0, cdb, 10, NULL, 0, 0, SCSI_IO_DIR_NONE, NULL, NULL);
	assert(ret == 0);
}

/* currently no real SCSI driver */
static struct driver *
fake_scsi_init(void)
{
	static struct driver fake_scsi_driver;
	return &fake_scsi_driver;
}

struct driver *dp83820_init(void);
struct driver *tigon3_init(void);
struct driver *ide_init(void);

#define PCIID(vendor, device) ((device << 16) | vendor)
#define NDEVICES (sizeof(devices) / sizeof(devices[0]))

struct device devices[] = {
#ifdef TIGON3
	{ PCIID(0x14e4, 0x1645), tigon3_init,    pci_net_setup,  "tigon3" },
#else
	{ PCIID(0x100b, 0x0022), dp83820_init,   pci_net_setup,  "dp83820" },
#endif
	{ PCIID(0x1095, 0x0680), ide_init,       pci_disk_setup, "cmd680" },
	{ PCIID(0x8086, 0x7601), ide_init,       pci_disk_setup, "piix5" },
	{ PCIID(0x1000, 0x0030), fake_scsi_init, pci_scsi_setup, "mpt" }
};

extern ia64_sal_handler sal;

uint8_t
pci_read_config_byte(uintptr_t address)
{
	return sal(SAL_PCI_CONFIG_READ, address, 1, 0, 0, 0, 0, 0).v0;
}

void
pci_write_config_byte(uintptr_t address, uint8_t value)
{
	sal(SAL_PCI_CONFIG_WRITE, address, 1, value, 0, 0, 0, 0);
}

uint16_t
pci_read_config_word(uintptr_t address)
{
	return sal(SAL_PCI_CONFIG_READ, address, 2, 0, 0, 0, 0, 0).v0;
}

void
pci_write_config_word(uintptr_t address, uint16_t value)
{
	sal(SAL_PCI_CONFIG_WRITE, address, 2, value, 0, 0, 0, 0);
}

uint32_t
pci_read_config_dword(uintptr_t address)
{
	return sal(SAL_PCI_CONFIG_READ, address, 4, 0, 0, 0, 0, 0).v0;
}

void
pci_write_config_dword(uintptr_t address, uint32_t value)
{
	sal(SAL_PCI_CONFIG_WRITE, address, 4, value, 0, 0, 0, 0);
}

static unsigned int
AcpiFindGSI(ACPI_BUFFER *RoutingTable, unsigned int bus, unsigned int device, unsigned int pin);

static void
pci_scan_bus(unsigned int bus, ACPI_BUFFER *irq_routing_table)
{
	unsigned int device, function;
	uint32_t value;
	bus_space_t pciconf, space[5];
	bs_seq spaces;
	struct driver *driver;
	unsigned int i, multifunction, gsi = 0;
	unsigned char pin;
	int id;

	for (device = 0; device < 32; device++)
	{
		multifunction = 1;
		for (function = 0; multifunction && (function < 8); function++)
		{
			id = -1;
			pciconf.type = IORESOURCE_PCICONF;
			pciconf.address = (bus << 16) | (device << 11) | (function << 8);
			value = io_read_32(pciconf, 0);
			if (value != 0xffffffff)
			{
				for (i = 0; i < NDEVICES; i++)
				{
					if (devices[i].pciid == value)
					{
						id = i;
						break;
					}
				}

				pin = io_read_8(pciconf, 0x3d);
				gsi = AcpiFindGSI(irq_routing_table, bus, device, (pin > 0) ? pin-1 : 0);

				printf("PCI: %02x:%02x.%x %04x:%04x %02x %s\n", bus, device, function,
				       value & 0xffff, value >> 16, gsi, (id != -1) ? devices[id].name : "");

				if (function == 0)
				{
					value = io_read_8(pciconf, 0xe);
					if (!(value & (1 << 7)))
						multifunction = 0;
				}
			}
			else if (function == 0)
			{
				multifunction = 0;
				break;
			}

			if (id == -1)
				continue;

			for (i = 0; i < 5; i++)
			{
				value = io_read_32(pciconf, 0x10 + i*4);
				if (value & 1)
				{
					space[i].type = IORESOURCE_IO;
					space[i].address = value & ~1UL;
					printf("BAR%x: I/O %lx\n", i, space[i].address);
				}
				else
				{
					space[i].type = IORESOURCE_MEM;
					space[i].address = (value & ~0xfUL) | (1UL<<63);
					printf("BAR%x: Mem %lx\n", i, space[i].address);
				}
			}
			spaces.c = 5;
			spaces.v = space;

			driver = devices[id].driver_init();
			devices[id].class_setup(&devices[id], driver, spaces, pciconf, gsi);
		}
	}
}

int
setup_dma(dma_handle_t dma, mem_handle_t handle, size_t size, dmabuf_handle_t *dmabuf, direction_t dir)
{
	dmabuf->vaddr = handle;
	dmabuf->addr = handle;
	dmabuf->size = size;
	return 0;
}

int
teardown_dma(dma_handle_t dma, dmabuf_handle_t *dmabuf)
{
	return 0;
}

int
get_pci_consistent_memory(dma_handle_t dma, size_t size, mem_handle_t *vaddr, dmabuf_handle_t *dmabuf)
{
	*vaddr = (mem_handle_t)malloc(size);
	if (*vaddr == 0)
		return -1;
	return setup_dma(dma, *vaddr, size, dmabuf, consistent);
}

#define SECTOR_BITS 9
int
disk_xfer(struct scsi_sglist *sgl, unsigned long offset, unsigned int length, int rw, void (*callback)(void *, uint64_t))
{
	char cdb[10];
	uint32_t blockid = offset >> SECTOR_BITS;
	uint32_t count = length >> SECTOR_BITS;
	block_seq data;
	block_t block;
	int ret;

	if (scsi)
	{
		//printf("disk_xfer %p %x %x %c\n", sgl, blockid, count, rw ? 'w' : 'r');
		/* assert(!rw); */
		memset(cdb, 0, 10);
		cdb[0] = rw ? 0x2a : 0x28 /* WRITE_10 : READ_10 */;
		cdb[2] = (blockid >> 24) & 0xff;
		cdb[3] = (blockid >> 16) & 0xff;
		cdb[4] = (blockid >> 8) & 0xff;
		cdb[5] = blockid & 0xff;
		cdb[7] = (count >> 8) & 0xff;
		cdb[8] = count & 0xff;
		ret = mpt_scsi_io_request(scsi, 0, 0, cdb, 10, sgl->sge, sgl->nsge, length, rw ? SCSI_IO_DIR_OUT : SCSI_IO_DIR_IN, callback, sgl);
		if (ret != 0)
			printf("ret = %d\n", ret);
		return 0;
	}
#if 0 /* FIXME: scatter-gather */
	else if (disk)
	{
		block.handle = new_mem_handle(buffer, size);
		block.blockid = blockid;
		block.block_count = count;
		block.key = 1;
		data.c = 1;
		data.v = &block;
		if (rw)
			return blkdev_write(disk, 0, callback, callback_data, data);
		else
			return blkdev_read(disk, 0, callback, callback_data, data);
	}
#endif
	else
	{
		printf("disk_xfer: no disks available\n");
		return -1;
	}
}

extern unsigned long hcdp;

void
early_serial_setup(void)
{
	struct hcdp_table *hcdp_table = (struct hcdp_table *)hcdp;
	struct hcdp_dev *console = NULL;
	bus_space_t space;
	unsigned int i;

#ifdef ITANIUM2 /* FIXME */
	ia64_io_base = 0x3FFFFFC000000;
#else
	ia64_io_base = 0xFFFFC000000;
#endif
	ia64_io_base |= 1UL << 63;

	if (hcdp_table == NULL)
		goto defaults;

	for (i = 0; i < hcdp_table->num_entries; i++)
	{
		if (hcdp_table->hcdp_dev[i].type == HCDP_DEV_CONSOLE)
		{
			console = &hcdp_table->hcdp_dev[i];
			break;
		}
	}

	if (console == NULL)
		goto defaults;

	switch (console->base_addr.AddressSpaceId)
	{
	case ACPI_ADR_SPACE_SYSTEM_MEMORY:
		space.type = IORESOURCE_MEM;
		space.address = console->base_addr.Address | (1UL << 63);
		break;
	case ACPI_ADR_SPACE_SYSTEM_IO:
		space.type = IORESOURCE_IO;
		space.address = console->base_addr.Address;
		break;
	default:
		goto defaults;
	}
	serial_gsi = console->global_int;
	serial = uart16550_setup(space);
	printf("serial: using %s=%lx irq=%x (from HCDP)\n",
	       (space.type == IORESOURCE_MEM) ? "mem" : "io", space.address, serial_gsi);
	return;

 defaults:
	space.type = IORESOURCE_IO;
	space.address = 0x3f8;
	serial_gsi = 4;
	serial = uart16550_setup(space);
	printf("serial: using io=0x3f8 irq=3 (defaults)\n");
}

static int received_char;
extern int guest_serial_irq;

int
getc(void)
{
	int c = received_char;
	received_char = 0;
	return c;
}

static void
serial_receive_char(int c)
{
	static int escape = 0;

	if (escape && (c != '~'))
	{
		process_debug_char(c);
		escape = 0;
		return;
	}
	if (!escape && (c == '~'))
	{
		escape = 1;
		return;
	}

	if (guest_serial_irq)
	{
		received_char = c;
		deliver_interrupt(guest_serial_irq);
	}
}

static void
serial_setup(void)
{
	uart16550_enable_receive(serial, serial_receive_char);
	register_irq(serial_gsi, uart16550_interrupt, serial, "serial");
}

static void
timer_interrupt(void *arg)
{
	deliver_interrupt(cpu.cr[CR_ITV] & 0xff);
}

static void
timer_setup(void)
{
	interrupts[0xff].handler = timer_interrupt;
	write_itv(0xff);
}

static void
pmu_interrupt(void *arg)
{
#ifdef PMU_DIRECT
	unsigned long pmc0;
	unsigned int i;

	/* XXX in case monitoring is stopped - Linux may unfreeze PMU without clearing overflow */
	pmc0 = read_pmc(0);
	for (i = 4; i < 8; i++)
	{
		if (pmc0 & (1UL << i))
			write_pmd(i, 0);
	}

	deliver_interrupt(cpu.cr[CR_PMV] & 0xff);
#else
	emu_fwb();
	write_pmc(0, 0); /* unfreeze PMU */
#endif
}

static void
pmu_setup(void)
{
	interrupts[0xfe].handler = pmu_interrupt;
	write_pmv(0xfe);
}

void
AcpiError(char *context, ACPI_STATUS Status)
{
	UserPrint("ACPI: %s: %s\n", context, AcpiFormatException(Status));
}

static unsigned int
AcpiFindGSI(ACPI_BUFFER *RoutingTable, unsigned int bus, unsigned int device, unsigned int pin)
{
	ACPI_PCI_ROUTING_TABLE *Element;
	char *ElementP;

	for (ElementP = (char *)(RoutingTable->Pointer);
	     ElementP < (char *)(RoutingTable->Pointer) + RoutingTable->Length;
	     ElementP += Element->Length)
	{
		Element = (ACPI_PCI_ROUTING_TABLE *)ElementP;
		if (Element->Length == 0)
			break;

		if(((Element->Address >> 16) == device)
		    && (Element->Pin == pin))
			return Element->SourceIndex;
	}
	return 0;
}

static ACPI_STATUS
AcpiScanBus( ACPI_HANDLE Handle, UINT32 NestingLevel, void *Context, 
	     void **ReturnValue )
{
	ACPI_BUFFER Buffer;
	ACPI_STATUS Status;
	ACPI_INTEGER BusId;
	union acpi_object BusIdObj;

	Buffer.Length = sizeof(BusIdObj);
	Buffer.Pointer = &BusIdObj;
	Status = AcpiEvaluateObjectTyped(Handle, METHOD_NAME__BBN, NULL, 
					 &Buffer, ACPI_TYPE_INTEGER);
	BusId = ACPI_FAILURE(Status) ? 0 : BusIdObj.Integer.Value;

	Buffer.Length = ACPI_ALLOCATE_BUFFER;
	Buffer.Pointer = NULL;
	Status = AcpiGetIrqRoutingTable(Handle, &Buffer);
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiGetIrqRoutingTable", Status);
		return Status;
	}

	pci_scan_bus(BusId, &Buffer);
	AcpiOsFree(Buffer.Pointer);
	return AE_OK;
}

#define APIC_IO_S	6

typedef struct io_sapic
{
    APIC_HEADER             Header;
    UINT8                   IoApicId;               /* I/O APIC ID */
    UINT8                   Reserved;               /* Reserved - must be zero */
    UINT32                  Vector;                 /* Interrupt vector index where INTI
                                                     * lines start */
    UINT32                  IoApicAddressLow;       /* APIC's physical address */
    UINT32                  IoApicAddressHigh;

} __attribute__((packed)) IO_SAPIC;

static void
AcpiRouteIrqs(void)
{
	ACPI_BUFFER Buffer;
	ACPI_STATUS Status;
	IO_SAPIC *Element;
	char *ElementP;
	uintptr_t Address;
	unsigned int i, NumRte, irq;

	Buffer.Length = ACPI_ALLOCATE_BUFFER;
	Buffer.Pointer = NULL;
	Status = AcpiGetTable(ACPI_TABLE_APIC, 1, &Buffer);
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiGetTable", Status);
		return;
	}

	for (ElementP = (char *)(Buffer.Pointer) + sizeof(APIC_TABLE);
	     ElementP < (char *)(Buffer.Pointer) + Buffer.Length;
	     ElementP += Element->Header.Length)
	{
		Element = (IO_SAPIC *)ElementP;
		if (Element->Header.Length == 0)
			break;
		if (Element->Header.Type != APIC_IO_S)
			continue;

		Address = (UINT64)(Element->IoApicAddressLow) | ((UINT64)(Element->IoApicAddressHigh) << 32) | (1UL << 63);
		writel(Address + IOSAPIC_REG_SELECT, IOSAPIC_VERSION);
		NumRte = (readl(Address + IOSAPIC_WINDOW) >> 16) + 1;
		printf("ACPI: IOSAPIC @ %lx (%x-%x)\n", Address, Element->Vector, Element->Vector+NumRte-1);
		for (i = 0; i < NumRte; i++)
		{
			irq = Element->Vector + i + 16;
			assert(irq < 255);
			interrupts[irq].iosapic_addr = Address;
			interrupts[irq].rte = i;
			interrupts[irq].eoi = (irq >= 0x20);
		}
	}

	AcpiOsFree(Buffer.Pointer);
}

#define DEBUG(x) printf x

static int
AcpiInit(void)
{
	ACPI_STATUS Status;

#ifdef ACPI_DEBUG
	AcpiDbgLevel = ACPI_LV_ALL;
	AcpiDbgLayer = ACPI_ALL_COMPONENTS;
#endif

	DEBUG(("ACPI: Initialising subsystem\n"));
	Status = AcpiInitializeSubsystem();
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiInitializeSubsystem", Status);
		return 1;
	}

	DEBUG(("ACPI: Loading tables\n"));
	Status = AcpiLoadTables();
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiLoadTables", Status);
		goto error;
	}

	DEBUG(("ACPI: Switching hardware to ACPI mode\n"));
	Status = AcpiEnableSubsystem(ACPI_FULL_INITIALIZATION);
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiEnableSubsystem", Status);
		goto error;
	}

	DEBUG(("ACPI: Initialising objects\n"));
	Status = AcpiInitializeObjects(ACPI_FULL_INITIALIZATION);
	if (ACPI_FAILURE(Status))
	{
		AcpiError("AcpiInitializeObjects", Status);
		goto error;
	}
	return 0;

error:
	AcpiTerminate();
	return 1;
}

void
device_init(void)
{
	/* mask legacy 8529A interrupts, we later route these via the IOSAPIC */
	outb(0xa1, 0xff);
	outb(0x21, 0xff);

	/* use ACPI to set up PCI interrupt routing and find buses */
	AcpiInit();
	AcpiRouteIrqs();
	AcpiGetDevices(PCI_ROOT_HID_STRING, AcpiScanBus, NULL, NULL);

	/* set up timer, PMU and serial interrupts */
	timer_setup();
	pmu_setup();
	serial_setup();
}

void
device_enter_virtual(void)
{
	bus_space_t *space;

#if 1
	ia64_io_base = ia64_io_base & ~(1UL << 63);
	dtr_insert(ia64_io_base, ia64_io_base, 26, PTE_AR_RW|PTE_MA_UC);
#endif

#ifdef TIGON3
	/* FIXME... this is all very ugly */
	tigon3_enter_virtual(netif);
#else
	dp83820_enter_virtual(netif);
#endif

	space = (bus_space_t *)((void **)serial + 1); /* XXX HACK */
	if (space->type == IORESOURCE_MEM)
	{
		printf("Serial port entering virtual mode\n");
		space->address &= ~(1UL << 63);
		dtr_insert(space->address, space->address, 12, PTE_AR_RW|PTE_MA_UC);
	}

	virtual_mode = 1;
}
